# IMPORT ROS
from typing import Any

import rospy
import rx.disposable
from rx import Observable, create
from rx.disposable import Disposable
from std_msgs.msg import UInt64, Bool

# IMPORT OTHER
from termcolor import cprint
import logging
import types
from functools import wraps
from threading import Condition

# IMPORT EAGERX
from eagerx.core.constants import DEBUG


def thread_safe_wrapper(func, condition):
    @wraps(func)
    def wrapped(*args, **kwargs):
        with condition:
            return func(*args, **kwargs)

    return wrapped


class RxMessageBroker(object):
    def __init__(self, owner):
        self.owner = owner

        # Determine log_level
        self.effective_log_level = logging.getLogger("rosout").getEffectiveLevel()

        # Ensure that we are not reading and writing at the same time.
        self.cond = Condition()

        # Structured as outputs[address][node_name] = {rx=Subject, node_name=node_name, source=RxOutput(...), etc..}
        self.rx_connectable = dict()

        # Structured as node_io[node_name][type][address] = {rx=Subject, disposable=rx_disposable, etc..}
        self.node_io = dict()
        self.disconnected = dict()
        self.connected_ros = dict()
        self.connected_rx = dict()

        # All publishers and subscribers (grouped to unregister when shutting down)
        self._publishers = []
        self.subscribers = []
        self.disposables = []

    # Every method is wrapped in a 'with Condition' block in order to be threadsafe
    def __getattribute__(self, name):
        attr = super(RxMessageBroker, self).__getattribute__(name)
        if isinstance(attr, types.MethodType):
            attr = thread_safe_wrapper(attr, self.cond)
        return attr

    def add_rx_objects(
        self,
        node_name,
        node=None,
        inputs=tuple(),
        outputs=tuple(),
        feedthrough=tuple(),
        state_inputs=tuple(),
        state_outputs=tuple(),
        targets=tuple(),
        node_inputs=tuple(),
        node_outputs=tuple(),
        reactive_proxy=tuple(),
        disposable: rx.disposable.CompositeDisposable = None,
    ):
        # Add disposable
        if disposable:
            self.disposables.append(disposable)

        # Determine tick address
        if node:
            ns = node.ns
        else:
            ns = self.node_io[node_name]["node"].ns
        tick_address = ns + "/bridge/outputs/tick"

        # Only add outputs that we would like to link with rx (i.e., skipping ROS (de)serialization)
        for i in outputs:
            if i["address"] == tick_address:
                continue
            assert i["address"] not in self.rx_connectable, (
                "Non-unique output (%s). All output names must be unique." % i["address"]
            )
            self.rx_connectable[i["address"]] = dict(rx=i["msg"], source=i, node_name=node_name, rate=i["rate"])

        # Register all I/O of node
        if node_name not in self.node_io:
            assert node is not None, 'No reference to Node "%s" was provided, during the first attempt to register it.'
            # Prepare io dictionaries
            self.node_io[node_name] = dict(
                node=node,
                inputs={},
                outputs={},
                feedthrough={},
                state_inputs={},
                state_outputs={},
                targets={},
                node_inputs={},
                node_outputs={},
                reactive_proxy={},
            )
            self.disconnected[node_name] = dict(inputs={}, feedthrough={}, state_inputs={}, targets={}, node_inputs={})
            self.connected_ros[node_name] = dict(inputs={}, feedthrough={}, state_inputs={}, targets={}, node_inputs={})
            self.connected_rx[node_name] = dict(inputs={}, feedthrough={}, state_inputs={}, targets={}, node_inputs={})
        n = dict(
            inputs={},
            outputs={},
            feedthrough={},
            state_inputs={},
            state_outputs={},
            targets={},
            node_inputs={},
            node_outputs={},
            reactive_proxy={},
        )
        for i in inputs:
            address = i["address"]
            cname_address = f"{i['name']}:{address}"
            self._assert_already_registered(cname_address, self.node_io[node_name], "inputs")
            self._assert_already_registered(cname_address, n, "inputs")
            n["inputs"][cname_address] = {
                "rx": i["msg"],
                "disposable": None,
                "source": i,
                "msg_type": i["msg_type"],
                "converter": i["converter"],
                "window": i["window"],
                "status": "disconnected",
            }
            n["inputs"][cname_address + "/reset"] = {
                "rx": i["reset"],
                "disposable": None,
                "source": i,
                "msg_type": UInt64,
                "status": "disconnected",
            }
        for i in outputs:
            address = i["address"]
            cname_address = f"{i['name']}:{address}"
            self._assert_already_registered(cname_address, self.node_io[node_name], "outputs")
            self._assert_already_registered(cname_address, n, "outputs")
            n["outputs"][cname_address] = {
                "rx": i["msg"],
                "disposable": None,
                "source": i,
                "msg_type": i["msg_type"],
                "rate": i["rate"],
                "converter": i["converter"],
                "status": "",
            }
            n["outputs"][cname_address + "/reset"] = {
                "rx": i["reset"],
                "disposable": None,
                "source": i,
                "msg_type": UInt64,
                "status": "",
            }

            # Create publisher
            i["msg_pub"] = rospy.Publisher(i["address"], i["msg_type"], queue_size=0, latch=True)
            d = i["msg"].subscribe(
                on_next=i["msg_pub"].publish,
                on_error=lambda e: print("Error : {0}".format(e)),
            )
            self.disposables.append(d)
            self._publishers.append(i["msg_pub"])
            i["reset_pub"] = rospy.Publisher(i["address"] + "/reset", UInt64, queue_size=0, latch=True)
            d = i["reset"].subscribe(
                on_next=i["reset_pub"].publish,
                on_error=lambda e: print("Error : {0}".format(e)),
            )
            self.disposables.append(d)
            self._publishers.append(i["reset_pub"])
        for i in feedthrough:
            address = i["address"]
            cname_address = f"{i['feedthrough_to']}:{address}"
            self._assert_already_registered(cname_address, self.node_io[node_name], "feedthrough")
            self._assert_already_registered(cname_address, n, "feedthrough")
            n["feedthrough"][cname_address] = {
                "rx": i["msg"],
                "disposable": None,
                "source": i,
                "msg_type": i["msg_type"],
                "converter": i["converter"],
                "window": i["window"],
                "status": "disconnected",
            }
            n["feedthrough"][cname_address + "/reset"] = {
                "rx": i["reset"],
                "disposable": None,
                "source": i,
                "msg_type": UInt64,
                "status": "disconnected",
            }
        for i in state_outputs:
            address = i["address"]
            cname_address = f"{i['name']}:{address}"
            self._assert_already_registered(cname_address, self.node_io[node_name], "state_outputs")
            self._assert_already_registered(cname_address, n, "state_outputs")
            n["state_outputs"][cname_address] = {
                "rx": i["msg"],
                "disposable": None,
                "source": i,
                "msg_type": i["msg_type"],
                "status": "",
            }
            if "converter" in i:
                n["state_outputs"][cname_address]["converter"] = i["converter"]

            # Create publisher
            i["msg_pub"] = rospy.Publisher(i["address"], i["msg_type"], queue_size=0, latch=True)
            d = i["msg"].subscribe(
                on_next=i["msg_pub"].publish,
                on_error=lambda e: print("Error : {0}".format(e)),
            )
            self.disposables.append(d)
            self._publishers.append(i["msg_pub"])
        for i in state_inputs:
            address = i["address"]
            try:
                cname_address = f"{i['name']}:{address}"
            except KeyError:
                cname_address = f"done_flag:{address}"
            if "msg" in i:  # Only true if sim state node (i.e. **not** for bridge done flags)
                self._assert_already_registered(cname_address + "/set", self.node_io[node_name], "state_inputs")
                self._assert_already_registered(cname_address + "/set", n, "state_inputs")
                n["state_inputs"][cname_address + "/set"] = {
                    "rx": i["msg"],
                    "disposable": None,
                    "source": i,
                    "msg_type": i["msg_type"],
                    "converter": i["converter"],
                    "status": "disconnected",
                }
            # Only true if **not** a real reset node (i.e., sim state & bridge done flag)
            if (cname_address + "/done") not in n["state_outputs"].keys():
                self._assert_already_registered(cname_address + "/done", self.node_io[node_name], "state_inputs")
                self._assert_already_registered(cname_address + "/done", n, "state_inputs")
                n["state_inputs"][cname_address + "/done"] = {
                    "rx": i["done"],
                    "disposable": None,
                    "source": i,
                    "msg_type": Bool,
                    "status": "disconnected",
                }
        for i in targets:
            address = i["address"]
            cname_address = f"{i['name']}:{address}"
            self._assert_already_registered(cname_address + "/set", self.node_io[node_name], "targets")
            self._assert_already_registered(cname_address + "/set", n, "targets")
            n["targets"][cname_address + "/set"] = {
                "rx": i["msg"],
                "disposable": None,
                "source": i,
                "msg_type": i["msg_type"],
                "converter": i["converter"],
                "status": "disconnected",
            }
        for i in node_inputs:
            address = i["address"]
            cname_address = f"{i['name']}:{address}"
            self._assert_already_registered(cname_address, self.node_io[node_name], "node_inputs")
            self._assert_already_registered(cname_address, n, "node_inputs")
            n["node_inputs"][cname_address] = {
                "rx": i["msg"],
                "disposable": None,
                "source": i,
                "msg_type": i["msg_type"],
                "status": "disconnected",
            }
        for i in node_outputs:
            address = i["address"]
            cname_address = f"{i['name']}:{address}"
            self._assert_already_registered(cname_address, self.node_io[node_name], "node_outputs")
            self._assert_already_registered(cname_address, n, "node_outputs")
            n["node_outputs"][cname_address] = {
                "rx": i["msg"],
                "disposable": None,
                "source": i,
                "msg_type": i["msg_type"],
                "status": "",
            }

            # Create publisher: (latched: register, node_reset, start_reset, reset, real_reset)
            i["msg_pub"] = rospy.Publisher(i["address"], i["msg_type"], queue_size=0, latch=True)
            d = i["msg"].subscribe(
                on_next=i["msg_pub"].publish,
                on_error=lambda e: print("Error : {0}".format(e)),
            )
            self.disposables.append(d)
            self._publishers.append(i["msg_pub"])
        for i in reactive_proxy:
            address = i["address"]
            cname_address = f"{i['name']}:{address}"
            self._assert_already_registered(cname_address, self.node_io[node_name], "reactive_proxy")
            self._assert_already_registered(cname_address, n, "reactive_proxy")
            n["reactive_proxy"][cname_address + "/reset"] = {
                "rx": i["reset"],
                "disposable": None,
                "source": i,
                "msg_type": UInt64,
                "rate": i["external_rate"],
                "status": "",
            }

            # Create publisher
            i["reset_pub"] = rospy.Publisher(i["address"] + "/reset", UInt64, queue_size=0, latch=True)
            d = i["reset"].subscribe(
                on_next=i["reset_pub"].publish,
                on_error=lambda e: print("Error : {0}".format(e)),
            )
            self.disposables.append(d)
            self._publishers.append(i["reset_pub"])

        # Add new addresses to already registered I/Os
        for key in n.keys():
            self.node_io[node_name][key].update(n[key])

        # Add new addresses to disconnected
        for key in ("inputs", "feedthrough", "state_inputs", "targets", "node_inputs"):
            self.disconnected[node_name][key].update(n[key].copy())

    def print_io_status(self, node_names=None):
        # Only print status for specific node
        if node_names is None:
            node_names = self.node_io.keys()
        else:
            if isinstance(node_names, str):
                node_names = [node_names]

        # Print status
        for node_name in node_names:
            cprint(
                ('OWNER "%s"' % self.owner).ljust(15, " ") + ('| OVERVIEW NODE "%s" ' % node_name).ljust(180, " "),
                attrs=["bold", "underline"],
            )
            for key in (
                "inputs",
                "feedthrough",
                "state_inputs",
                "targets",
                "node_inputs",
                "outputs",
                "state_outputs",
                "node_outputs",
                "reactive_proxy",
            ):
                if len(self.node_io[node_name][key]) == 0:
                    continue
                for cname_address in self.node_io[node_name][key].keys():
                    color = None
                    if key in (
                        "outputs",
                        "node_outputs",
                        "state_outputs",
                        "reactive_proxy",
                    ):
                        color = "cyan"
                    else:
                        if cname_address in self.disconnected[node_name][key]:
                            color = "red"
                        if cname_address in self.connected_rx[node_name][key]:
                            assert color is None, f"Duplicate connection status for address ({cname_address})."
                            color = "green"
                        if cname_address in self.connected_ros[node_name][key]:
                            assert color is None, f"Duplicate connection status for address ({cname_address})."
                            color = "blue"
                        assert (
                            color is not None
                        ), "Address (cname_address) not found in self.(disconnected, connected_rx, connected_ros)."
                    status = self.node_io[node_name][key][cname_address]["status"]

                    # Print status
                    entry = self.node_io[node_name][key][cname_address]
                    key_str = ("%s" % key).ljust(15, " ")
                    address_str = ("| %s " % cname_address).ljust(50, " ")
                    msg_type_str = ("| %s " % entry["msg_type"].__name__).ljust(10, " ")
                    if "converter" in entry:
                        converter_str = ("| %s " % entry["converter"].__class__.__name__).ljust(23, " ")
                    else:
                        converter_str = ("| %s " % "").ljust(23, " ")
                    if "window" in entry:
                        window_str = ("| %s " % entry["window"]).ljust(8, " ")
                    else:
                        window_str = ("| %s " % "").ljust(8, " ")
                    if "rate" in entry:
                        rate_str = "|" + ("%s" % entry["rate"]).center(3, " ")
                    else:
                        rate_str = "|" + "".center(3, " ")
                    status_str = ("| %s" % status).ljust(60, " ")

                    log_msg = key_str + rate_str + address_str + msg_type_str + converter_str + window_str + status_str
                    cprint(log_msg, color)
            print(" ".center(140, " "))

    def connect_io(self, print_status=True):
        # If log_level is not high enough, overwrite print_status
        if self.effective_log_level > DEBUG:
            print_status = False

        for node_name, node in self.disconnected.items():
            # Skip if no disconnected addresses
            num_disconnected = 0

            for _key, addresses in node.items():
                num_disconnected += len(addresses)
            if num_disconnected == 0:
                continue

            # Else, initialize connection
            print_status and cprint(
                ('OWNER "%s"' % self.owner).ljust(15, " ") + ('| CONNECTING NODE "%s" ' % node_name).ljust(180, " "),
                attrs=["bold", "underline"],
            )
            for key, addresses in node.items():
                for cname_address in list(addresses.keys()):
                    _, address = self._split_cname_address(cname_address)
                    entry = addresses[cname_address]
                    assert (
                        cname_address not in self.connected_rx[node_name][key]
                    ), f"Address ({cname_address}) of this node ({node_name}) already connected via rx."
                    assert (
                        cname_address not in self.connected_ros[node_name][key]
                    ), f"Address ({cname_address}) of this node ({node_name}) already connected via ROS."
                    if address in self.rx_connectable.keys():
                        color = "green"
                        status = "Rx".ljust(4, " ")
                        entry["rate"] = self.rx_connectable[address]["rate"]
                        rate_str = f"|{str(entry['rate']).center(3, ' ')}"
                        node_str = f'| {self.rx_connectable[address]["node_name"].ljust(40, " ")}'
                        msg_type_str = f'| {self.rx_connectable[address]["source"]["msg_type"].__name__}'.ljust(12, " ")
                        converter_str = f'| {self.rx_connectable[address]["source"]["converter"].__class__.__name__}'.ljust(
                            12, " "
                        )
                        status += node_str + msg_type_str + converter_str
                        self.connected_rx[node_name][key][cname_address] = entry
                        T = self.rx_connectable[address]["rx"]
                    else:
                        color = "blue"
                        status = "ROS |".ljust(5, " ")
                        rate_str = "|" + "".center(3, " ")
                        msg_type = entry["msg_type"]
                        self.connected_ros[node_name][key][cname_address] = entry
                        T = from_topic(msg_type, address, node_name, self.subscribers)

                    # Subscribe and change status
                    entry["disposable"] = T.subscribe(entry["rx"])
                    self.disposables.append(entry["disposable"])
                    entry["status"] = status

                    # Print status
                    key_str = ("%s" % key).ljust(15, " ")
                    address_str = ("| %s " % cname_address).ljust(50, " ")
                    msg_type_str = ("| %s " % entry["msg_type"].__name__).ljust(10, " ")
                    status_str = ("| Connected via %s" % status).ljust(60, " ")

                    if "converter" in entry:
                        converter_str = ("| %s " % entry["converter"].__class__.__name__).ljust(23, " ")
                    else:
                        converter_str = ("| %s " % "").ljust(23, " ")
                    if "window" in entry:
                        window_str = ("| %s " % entry["window"]).ljust(8, " ")
                    else:
                        window_str = ("| %s " % "").ljust(8, " ")

                    log_msg = key_str + rate_str + address_str + msg_type_str + converter_str + window_str + status_str
                    print_status and cprint(log_msg, color)

                    # Remove address from disconnected
                    addresses.pop(cname_address)

            print_status and print("".center(140, " "))

    def _split_cname_address(self, cname_address):
        res = cname_address.split(":")
        if len(res) == 2:
            cname, address = res
        else:
            cname, address = None, res[0]
        return cname, address

    def _assert_already_registered(self, name, d, component):
        assert name not in d[component], f'Cannot re-register the same address ({name}) twice as "{component}".'

    def shutdown(self):
        rospy.logdebug(f"[{self.owner}] RxMessageBroker.shutdown() called.")
        [pub.unregister() for pub in self._publishers]
        [sub.unregister() for sub in self.subscribers]
        [d.dispose() for d in self.disposables]


def from_topic(topic_type: Any, topic_name: str, node_name, subscribers: list) -> Observable:
    def _subscribe(observer, scheduler=None) -> Disposable:
        try:
            wrapped_sub = []

            def cb_from_topic(msg, wrapped_sub):
                try:
                    observer.on_next(msg)
                except rospy.exceptions.ROSException as e:
                    sub = wrapped_sub[0]
                    sub.unregister()
                    rospy.logdebug(f"[{sub.name}]: Unregistered this subscription because of exception: {e}")

            sub = rospy.Subscriber(topic_name, topic_type, callback=cb_from_topic, callback_args=wrapped_sub)
            wrapped_sub.append(sub)
            subscribers.append(sub)
        except Exception as e:
            rospy.logwarn("[%s]: %s" % (node_name, e))
            raise
        return observer

    return create(_subscribe)



from typing import Dict, Any, Optional
import inspect
from yaml import dump
import copy

import eagerx.core.register as register
from eagerx.core.view import SpecView, GraphView
from eagerx.utils.utils import (
    replace_None,
    deepcopy,
    get_module_type_string,
    get_default_params,
    substitute_args,
)


class EntitySpec(object):
    def __init__(self, params):
        super(EntitySpec, self).__setattr__("_params", params)
        super(EntitySpec, self).__setattr__("_graph", None)

    def __setattr__(self, name, value):
        raise AttributeError("You cannot set the new attributes to EntitySpec.")

    def __str__(self):
        return dump(self._params)

    @property
    @deepcopy
    def params(self):
        return self._params

    def set_graph(self, graph):
        super(EntitySpec, self).__setattr__("_graph", graph)

    @property
    def has_graph(self):
        return True if self._graph else False

    @property
    def graph(self):
        return self._graph


class ConverterSpec(EntitySpec):
    def initialize(self, spec_cls):
        # Set default params
        defaults = get_default_params(spec_cls.initialize)
        with self.config as d:
            d.update(defaults)

    @property
    def config(self):
        return SpecView(self, depth=[])


class EngineStateSpec(EntitySpec):
    def initialize(self, spec_cls):
        # Set default params
        defaults = get_default_params(spec_cls.initialize)
        with self.config as d:
            d.update(defaults)

    @property
    def config(self):
        return SpecView(self, depth=[])


class BaseNodeSpec(EntitySpec):
    def __init__(self, params):
        super().__init__(params)
        from eagerx.core.converters import BaseConverter

        super(EntitySpec, self).__setattr__("identity", BaseConverter.make("Identity"))

    def _lookup(self, depth):
        name = self._params["config"]["name"]
        if self.has_graph:
            return GraphView(self.graph, depth=[name, depth], name=name)
        else:
            return SpecView(self, depth=[depth], name=name)

    @property
    def inputs(self):
        return self._lookup("inputs")

    @property
    def outputs(self):
        return self._lookup("outputs")

    @property
    def states(self):
        return self._lookup("states")

    @property
    def config(self):
        return self._lookup("config")

    def initialize(self, spec_cls):
        try:
            params = register.LOOKUP_TYPES[spec_cls.callback]
        except KeyError:
            if spec_cls.__name__ == "EnvNode":
                params = dict()
            else:
                raise

        # Set default params
        defaults = get_default_params(spec_cls.initialize)
        with self.config as d:
            d.update(defaults)

        if "bridge_config" in params:
            params.pop("bridge_config")

        if "targets" in params:
            from eagerx.core.entities import ResetNode

            assert issubclass(
                spec_cls, ResetNode
            ), "You can only have targets registered for nodes that inherit from the ResetNode baseclass."
            add_ft = True
        else:
            add_ft = False

        # Set default components
        for component, cnames in params.items():
            for cname, msg_type in cnames.items():
                msg_type = get_module_type_string(msg_type)
                if component == "outputs":
                    self.config.outputs.append(cname)
                    mapping = dict(
                        msg_type=msg_type,
                        rate="$(config rate)",
                        converter=self.identity.params,
                        space_converter=None,
                    )
                    # Add feedthrough entries for each output if node is a reset node (i.e. when it has a target)
                    if add_ft:
                        mapping_ft = dict(
                            msg_type=msg_type,
                            delay=0.0,
                            window=1,
                            skip=False,
                            external_rate=None,
                            converter=self.identity.params,
                            space_converter=None,
                            address=None,
                        )
                        with self.feedthroughs as d:
                            d[cname] = mapping_ft
                        # self._set({"feedthroughs": {cname: mapping_ft}})
                elif component == "inputs":
                    self.config.inputs.append(cname)
                    address = "bridge/outputs/tick" if cname == "tick" else None
                    mapping = dict(
                        msg_type=msg_type,
                        delay=0.0,
                        window=1,
                        skip=False,
                        external_rate=None,
                        converter=self.identity.params,
                        space_converter=None,
                        address=address,
                    )
                elif component == "targets":
                    self.config.targets.append(cname)
                    mapping = dict(
                        msg_type=msg_type,
                        converter=self.identity.params,
                        space_converter=None,
                        address=None,
                    )
                else:
                    self.config.states.append(cname)
                    component = "states"
                    mapping = dict(
                        msg_type=msg_type,
                        converter=self.identity.params,
                        space_converter=None,
                    )
                with getattr(self, component) as d:
                    d[cname] = mapping

    def _remove_component(self, component: str, cname: str):
        getattr(self, component).pop(cname)
        if cname in self.config[component]:
            self.config[component].remove(cname)

    def remove_input(self, cname: str):
        self._remove_component("inputs", cname)

    def remove_output(self, cname: str):
        self._remove_component("outputs", cname)

    def remove_state(self, cname: str):
        self._remove_component("states", cname)

    def remove_target(self, cname: str):
        self._remove_component("targets", cname)

    def add_input(
        self,
        cname: str,
        msg_type: Any,
        window: int = 1,
        delay: float = 0.0,
        skip: bool = False,
        external_rate: float = None,
        address: str = None,
        converter: Optional[ConverterSpec] = None,
        space_converter: Optional[ConverterSpec] = None,
    ):
        if not isinstance(msg_type, str):
            assert inspect.isclass(
                msg_type
            ), f'An instance "{msg_type}" of class "{msg_type.__class__}" was provided. Please provide the class instead.'
            msg_type = get_module_type_string(msg_type)
        mapping = dict(
            msg_type=msg_type,
            window=window,
            delay=delay,
            skip=skip,
            external_rate=external_rate,
            address=address,
        )
        mapping["converter"] = converter.params if converter else self.identity.params
        mapping["space_converter"] = space_converter.params if space_converter else None
        with self.inputs as d:
            d[cname] = mapping

    def add_output(
        self,
        cname: str,
        msg_type: Any,
        converter: Optional[ConverterSpec] = None,
        space_converter: Optional[ConverterSpec] = None,
    ):
        if not isinstance(msg_type, str):
            assert inspect.isclass(
                msg_type
            ), f'An instance "{msg_type}" of class "{msg_type.__class__}" was provided. Please provide the class instead.'
            msg_type = get_module_type_string(msg_type)
        mapping = dict(msg_type=msg_type, rate="$(config rate)")
        mapping["converter"] = converter.params if converter else self.identity.params
        mapping["space_converter"] = space_converter.params if space_converter else None
        with self.outputs as d:
            d[cname] = mapping

    def add_state(self, cname: str, msg_type: Any, space_converter: ConverterSpec):
        if not isinstance(msg_type, str):
            assert inspect.isclass(
                msg_type
            ), f'The provided msg_type "{msg_type}" is not a class. Make sure you are *not* providing an instance of the class, instead of the class itself.'
            msg_type = get_module_type_string(msg_type)
        mapping = dict(msg_type=msg_type, space_converter=space_converter.params)
        with self.states as d:
            d[cname] = mapping

    def add_target(self, cname: str, msg_type: Any, converter: Optional[ConverterSpec] = None):
        if not isinstance(msg_type, str):
            assert inspect.isclass(
                msg_type
            ), f'The provided msg_type "{msg_type}" is not a class. Make sure you are *not* providing an instance of the class, instead of the class itself.'
            msg_type = get_module_type_string(msg_type)
        mapping = dict(msg_type=msg_type)
        mapping["converter"] = converter.params if converter else self.identity.params

        assert "targets" in self._params["config"], f"Cannot add target '{cname}'. Node is not a 'ResetNode'"
        with self.targets as d:
            d[cname] = mapping

    def build(self, ns):
        params = self.params  # Creates a deepcopy
        default = copy.deepcopy(self.config.to_dict())
        name = default["name"]
        default["node_type"] = params["node_type"]
        entity_id = default["entity_id"]

        # Replace args in .yaml
        context = {
            "ns": {"env_name": ns, "node_name": name},
            "config": params["config"],
        }
        substitute_args(params, context, only=["config", "ns"])

        # Process inputs
        inputs = []
        for cname in default["inputs"]:
            assert (
                cname in params["inputs"]
            ), f'Received unknown {"input"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
            assert (
                "targets" not in params or cname not in params["targets"]
            ), f'Input "{cname}" cannot have the same cname as a target. Change either the input or target cname. Check the spec of "{name}" with entity_id "{entity_id}".'
            n = RxInput(name=cname, **params["inputs"][cname])
            inputs.append(n)

        # Process outputs
        outputs = []
        for cname in default["outputs"]:
            assert (
                cname in params["outputs"]
            ), f'Received unknown {"output"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
            if "address" in params["outputs"][cname]:
                address = params["outputs"][cname].pop("address")
            else:
                address = "%s/outputs/%s" % (name, cname)
            n = RxOutput(name=cname, address=address, **params["outputs"][cname])
            outputs.append(n)

        states = []
        for cname in default["states"]:
            assert (
                cname in params["states"]
            ), f'Received unknown {"state"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
            if "address" in params["states"][cname]:  # if 'env/supervisor', the state address is pre-defined (like an input)
                n = RxState(name=cname, **params["states"][cname])
            else:
                address = "%s/states/%s" % (name, cname)
                n = RxState(name=cname, address=address, **params["states"][cname])
            states.append(n)

        targets = []
        if "targets" in default:
            for cname in default["targets"]:
                assert (
                    cname in params["targets"]
                ), f'Received unknown {"target"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
                n = RxState(name=cname, **params["targets"][cname])
                targets.append(n)

        feedthroughs = []
        if "feedthroughs" in params:
            for cname in default["outputs"]:
                # Add output details (msg_type, space_converter) to feedthroughs
                assert (
                    cname in params["feedthroughs"]
                ), f'Feedthrough "{cname}" must directly correspond to a selected output. Check the spec of "{name}" with entity_id "{entity_id}".'
                assert (
                    params["outputs"][cname]["msg_type"] == params["feedthroughs"][cname]["msg_type"]
                ), f'Mismatch between Msg types of feedthrough "{cname}" and output "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
                if "space_converter" in params["outputs"][cname]:
                    params["feedthroughs"][cname]["space_converter"] = params["outputs"][cname]["space_converter"]
                n = RxFeedthrough(feedthrough_to=cname, **params["feedthroughs"][cname])
                feedthroughs.append(n)

        default["outputs"] = [i.build(ns=ns) for i in outputs]
        default["inputs"] = [i.build(ns=ns) for i in inputs]
        default["states"] = [i.build(ns=ns) for i in states]
        default["targets"] = [i.build(ns=ns) for i in targets]
        default["feedthroughs"] = [i.build(ns=ns) for i in feedthroughs]

        # Create rate dictionary with outputs
        chars_ns = len(ns) + 1
        rate_dict = dict()
        for i in default["outputs"]:
            assert (
                i["rate"] is not None and isinstance(i["rate"], (int, float)) and i["rate"] > 0
            ), f'The rate of node "{name}" (and output cname "{i["name"]}") is misspecified: rate="{i["rate"]}". Make sure that it is of type(rate)=("int", "float",) and rate > 0.'
            address = i["address"][chars_ns:]
            rate_dict[address] = i["rate"]  # {'rate': i['rate']}

        # Put parameters in node namespace (watch out, order of dict keys probably matters...)
        node_params = {name: default, "rate": rate_dict}
        return replace_None(node_params)


class NodeSpec(BaseNodeSpec):
    pass


class ResetNodeSpec(BaseNodeSpec):
    @property
    def targets(self):
        return self._lookup("targets")

    @property
    def feedthroughs(self):
        return self._lookup("feedthroughs")


class BridgeSpec(BaseNodeSpec):
    pass


class ObjectSpec(EntitySpec):
    def __init__(self, params):
        super().__init__(params)
        from eagerx.core.converters import BaseConverter

        super(EntitySpec, self).__setattr__("identity", BaseConverter.make("Identity"))

    def __getattr__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name in self._params:
                return SpecView(self, depth=[name], name=self._params["config"]["name"])
            else:
                raise

    def _lookup(self, depth):
        name = self._params["config"]["name"]
        if self.has_graph:
            return GraphView(self.graph, depth=[name, depth], name=name)
        else:
            return SpecView(self, depth=[depth], name=name)

    @property
    def sensors(self):
        return self._lookup("sensors")

    @property
    def actuators(self):
        return self._lookup("actuators")

    @property
    def states(self):
        return self._lookup("states")

    @property
    def config(self):
        return self._lookup("config")

    def initialize(self, spec_cls):
        agnostic = register.LOOKUP_TYPES[spec_cls.agnostic]

        # Set default agnostic params
        with self.config as d:
            d.update(agnostic.pop("config"))

        # Set default components
        for component, cnames in agnostic.items():
            for cname, msg_type in cnames.items():
                msg_type = get_module_type_string(msg_type)
                if component == "sensors":
                    mapping = dict(
                        msg_type=msg_type,
                        rate=1,
                        converter=self.identity.params,
                        space_converter=None,
                    )
                elif component == "actuators":
                    mapping = dict(
                        msg_type=msg_type,
                        rate=1,
                        delay=0.0,
                        window=1,
                        skip=False,
                        external_rate=None,
                        converter=self.identity.params,
                        space_converter=None,
                    )
                else:
                    component = "states"
                    mapping = dict(
                        msg_type=msg_type,
                        converter=self.identity.params,
                        space_converter=None,
                    )
                with getattr(self, component) as d:
                    d[cname] = mapping

    def _initialize_bridge_config(self, bridge_id, bridge_config):
        # Add default config
        with getattr(self, bridge_id) as d:
            d.update(bridge_config)
            d["states"] = {}
            # Add all states to engine-specific params
            with d.states as s:
                for cname in self.states.keys():
                    s[cname] = None

    def _add_graph(self, bridge_id, graph):
        # Register EngineGraph
        nodes, actuators, sensors = graph.register()

        # Pop states that were not implemented.
        with getattr(self, bridge_id).states as d:
            for cname in list(getattr(self, bridge_id).states.keys()):
                if d[cname] is None:
                    d.pop(cname)

        # Set engine_spec
        with getattr(self, bridge_id) as d:
            d.actuators = actuators
            d.sensors = sensors
            d.nodes = nodes

    def _initialize_object_graph(self):
        mapping = dict()
        for component in ["sensors", "actuators"]:
            try:
                mapping[component] = getattr(self, component)
            except AttributeError:
                continue

        from eagerx.core.graph_engine import EngineGraph

        graph = EngineGraph.create(**mapping)
        return graph

    def add_bridge(self, bridge_id):
        # Construct context & replace placeholders
        context = {"config": self.config.to_dict()}
        substitute_args(self._params["config"], context, only=["config"])  # First resolve args within the context
        substitute_args(self._params, context, only=["config"])  # Resolve rest of params

        # Add bridge entry
        self._params[bridge_id] = {}
        register.add_bridge(self, bridge_id)

    def build(self, ns, bridge_id):
        params = self.params  # Creates a deepcopy
        default = copy.deepcopy(self.config.to_dict())  # Creates a deepcopy
        name = default["name"]

        # Construct context
        context = {"ns": {"env_name": ns, "obj_name": name}}
        substitute_args(default, context, only=["ns"])  # First resolve args within the context
        substitute_args(params, context, only=["ns"])  # Resolve rest of params

        # Get agnostic definition
        agnostic = dict()
        for key in list(params.keys()):
            if key not in ["actuators", "sensors", "states"]:
                if key not in ["config", bridge_id]:
                    params.pop(key)
                continue
            agnostic[key] = params.pop(key)

        # Get bridge definition
        bridge = params.pop(bridge_id)
        nodes = bridge.pop("nodes")
        specific = dict()
        for key in list(bridge.keys()):
            if key not in ["actuators", "sensors", "states"]:
                continue
            specific[key] = bridge.pop(key)

        # Replace node names
        for key in list(nodes.keys()):
            key_sub = substitute_args(key, context, only=["ns"])
            nodes[key_sub] = nodes.pop(key)

        # Sensors & actuators
        sensor_addresses = dict()
        dependencies = []
        for obj_comp in ["actuators", "sensors"]:
            for obj_cname in default[obj_comp]:
                try:
                    entry_lst = specific[obj_comp][obj_cname]
                except KeyError:
                    raise KeyError(
                        f'"{obj_cname}" was selected in {obj_comp} of "{name}", but there is no implementation for it in bridge "{bridge_id}".'
                    )
                # todo: here we assume a single node implements the actuator --> could be multiple

                # entry = entry_lst
                for entry in reversed(entry_lst):
                    node_name, node_comp, node_cname = entry["name"], entry["component"], entry["cname"]
                    obj_comp_params = agnostic[obj_comp][obj_cname]
                    node_params = nodes[node_name]

                    # Determine node dependency
                    dependencies += entry["dependency"]

                    # Set rate
                    rate = obj_comp_params["rate"]
                    msg_start = f'Different rate specified for {obj_comp[:-1]} "{obj_cname}" and enginenode "{node_name}": '
                    msg_end = "If an enginenode implements a sensor/actuator, their specified rates must be equal."
                    msg_mid = f'{node_params["config"]["rate"]} vs {rate}. '
                    assert node_params["config"]["rate"] == rate, msg_start + msg_mid + msg_end
                    for o in node_params["config"]["outputs"]:
                        msg_mid = f'{node_params["outputs"][o]["rate"]} vs {rate}. '
                        assert node_params["outputs"][o]["rate"] == rate, msg_start + msg_mid + msg_end

                    # Set component params
                    node_comp_params = nodes[node_name][node_comp][node_cname]
                    if obj_comp == "sensors":
                        node_comp_params.update(obj_comp_params)
                        node_comp_params["address"] = f"{name}/{obj_comp}/{obj_cname}"
                        sensor_addresses[f"{node_name}/{node_comp}/{node_cname}"] = f"{name}/{obj_comp}/{obj_cname}"
                    else:  # Actuators
                        node_comp_params.update(obj_comp_params)
                        node_comp_params.pop("rate")

        # Get set of node we are required to launch
        dependencies = list(set(dependencies))

        # Verify that no dependency is an unlisted actuator node.
        not_selected = [cname for cname in agnostic["actuators"] if cname not in default["actuators"]]
        for cname in not_selected:
            try:
                for entry in specific["actuators"][cname]:
                    node_name, node_comp, node_cname = (entry["name"], entry["component"], entry["cname"])
                    msg = (
                        f'There appears to be a dependency on enginenode "{node_name}" for the implementation of '
                        f'bridge "{bridge_id}" for object "{name}" to work. However, enginenode "{node_name}" is '
                        f'directly tied to an unselected actuator "{cname}". '
                        "The actuator must be selected to resolve the graph."
                    )
                    assert node_name not in dependencies, msg
            except KeyError:
                # We pass here, because if cname is not selected, but also not implemented,
                # we are sure that there is no dependency.
                pass

        # Replace enginenode outputs that have been renamed to sensor outputs
        for node_address, sensor_address in sensor_addresses.items():
            for _, node_params in nodes.items():
                for _cname, comp_params in node_params["inputs"].items():
                    if node_address == comp_params["address"]:
                        comp_params["address"] = sensor_address

        # Create states
        states = []
        state_names = []
        obj_comp = "states"
        for obj_cname in default["states"]:
            args = agnostic[obj_comp][obj_cname]
            args["name"] = f"{name}/{obj_comp}/{obj_cname}"
            args["address"] = f"{name}/{obj_comp}/{obj_cname}"
            try:
                args["state"] = specific[obj_comp][obj_cname]
            except KeyError:
                raise KeyError(
                    f'"{obj_cname}" was selected in {obj_comp} of "{name}", but there is no implementation for it in bridge "{bridge_id}".'
                )
            states.append(RxEngineState(**args))
            state_names.append(f'{ns}/{args["name"]}')

        # Create obj parameters
        obj_params = params["config"]

        # Gather node names
        obj_params["node_names"] = [f"{ns}/{node_name}" for node_name in list(nodes.keys()) if node_name in dependencies]
        obj_params["state_names"] = state_names

        # Add bridge
        obj_params["bridge"] = bridge

        # Clean up parameters
        for component in ["sensors", "actuators", "states"]:
            try:
                obj_params.pop(component)
            except KeyError:
                pass

        # Add states
        obj_params["states"] = [s.build(ns) for s in states]
        nodes = [NodeSpec(params) for name, params in nodes.items() if name in dependencies]
        return {name: replace_None(obj_params)}, nodes


# REQUIRED FOR BUILDING SPECS


class Component(object):
    def __init__(self, **kwargs):
        # Iterates over provided arguments and sets the provided arguments as class properties
        for key, value in kwargs.items():
            if key == "__class__":
                continue  # Skip if __class__ type
            setattr(self, key, value)


class RxInput(Component):
    def __init__(
        self,
        name: str,
        address: str,
        msg_type: str,
        window: int = 0,
        converter: Dict = None,
        external_rate: float = None,
        space_converter: Dict = None,
        delay: float = 0.0,
        skip: bool = False,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxInput, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        if not params["external_rate"]:
            params["address"] = "/".join(filter(None, [ns, params["address"]]))
        return params


class RxOutput(Component):
    def __init__(
        self,
        name: str,
        address: str,
        msg_type: str,
        rate: float,
        converter: Dict = None,
        space_converter: Dict = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        # If space_converter undefined, remove it
        # if not space_converter:
        #     raise NotImplementedError(name)
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxOutput, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        return params


class RxFeedthrough(Component):
    def __init__(
        self,
        address: str,
        msg_type: str,
        feedthrough_to: str,
        window: int = 1,
        converter: Dict = None,
        external_rate: float = None,
        space_converter: Dict = None,
        delay: float = 0.0,
        skip: bool = False,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxFeedthrough, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        return params


class RxState(Component):
    def __init__(
        self,
        name: str,
        address: str,
        msg_type: str,
        converter: Dict = None,
        space_converter: Dict = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxState, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        return params


class RxEngineState(Component):
    def __init__(
        self,
        name: str,
        address: str,
        state: Dict,
        msg_type: str,
        converter: Dict = None,
        space_converter: Dict = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxEngineState, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        return params



# ROS IMPORTS
import rospy
from std_msgs.msg import Bool, UInt64

# RX IMPORTS
import rx
from rx import Observable, typing, operators as ops
from rx.disposable import Disposable, SingleAssignmentDisposable, CompositeDisposable
from rx.subject import Subject, BehaviorSubject
from rx.internal.concurrency import synchronized

# EAGERX IMPORTS
from eagerx.core.converters import Identity
from eagerx.core.specs import RxInput
from eagerx.core.constants import (
    SILENT,
    DEBUG,
    INFO,
    ERROR,
    WARN,
    FATAL,
    TERMCOLOR,
    ROS,
)
from eagerx.utils.utils import (
    get_attribute_from_module,
    initialize_converter,
    get_param_with_blocking,
    Info,
    Msg,
    Stamp,
    get_opposite_msg_cls,
    get_module_type_string,
    msg_type_error,
)

# OTHER IMPORTS
import time
from collections import deque
from termcolor import cprint
import datetime
import logging
import traceback
from os import getpid
from threading import current_thread, RLock
from typing import Callable, Optional, List, Any

print_modes = {TERMCOLOR: "eagerx_core.TERMCOLOR", ROS: "eagerx_core.ROS"}
ros_log_fns = {
    SILENT: lambda print_str: None,
    DEBUG: rospy.logdebug,
    INFO: rospy.loginfo,
    WARN: rospy.logwarn,
    ERROR: rospy.logerr,
    FATAL: rospy.logfatal,
}


def cb_ft(cb_input, is_reactive):
    # Fill output msg with number of node ticks
    output_msgs = dict()
    for key, msg in cb_input.items():
        if key not in ["node_tick", "t_n"]:
            if len(msg.msgs) > 0:
                output_msgs[key] = msg.msgs[-1]
            else:
                assert not is_reactive, "Actions must always be fed through if we are running reactively."
                output_msgs[key] = None
    return output_msgs


def print_info(
    node_name,
    color,
    id=None,
    trace_type=None,
    value=None,
    date=None,
    print_mode=TERMCOLOR,
    log_level=DEBUG,
):
    if print_mode == TERMCOLOR:
        if date:
            cprint("[" + str(date)[:40].ljust(20) + "]", color, end="")
        cprint("[" + str(getpid())[:5].ljust(5) + "]", color, end="")
        cprint(
            "[" + current_thread().name.split("/")[-1][:15].ljust(15) + "]",
            color,
            end="",
        )
        cprint(
            "[" + node_name.split("/")[-1][:12].ljust(12) + "]",
            color,
            end="",
            attrs=["bold"],
        )
        if id:
            cprint("[" + id.split("/")[-1][:12].ljust(12) + "]", color, end="")
        cprint((" %s: %s" % (trace_type, value)), color)
    elif print_mode == ROS:
        print_str = ""
        print_str += "[" + str(getpid())[:5].ljust(5) + "]"
        print_str += "[" + current_thread().name.split("/")[-1][:15].ljust(15) + "]"
        print_str += "[" + node_name.split("/")[-1][:12].ljust(12) + "]"
        if id:
            print_str += "[" + id.split("/")[-1][:12].ljust(12) + "]"
        print_str += " %s: %s" % (trace_type, value)
        ros_log_fns[log_level](print_str)
    else:
        raise ValueError("Print mode not recognized. Only print_modes %s are available." % (print_modes.values()))


def spy(id: str, node, log_level: int = DEBUG, mapper: Callable = lambda msg: msg):
    node_name = node.ns_name
    color = node.color
    print_mode = node.print_mode
    effective_log_level = logging.getLogger("rosout").getEffectiveLevel()

    def _spy(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):

                if node.log_level >= effective_log_level and log_level >= effective_log_level:
                    print_info(
                        node_name,
                        color,
                        id,
                        trace_type="",
                        value=str(mapper(value)),
                        print_mode=print_mode,
                        log_level=log_level,
                    )
                observer.on_next(value)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _spy


def trace_observable(
    id: str,
    node,
    trace_next=False,
    trace_next_payload=False,
    trace_subscribe=False,
    date=None,
):
    node_name = node.ns_name
    color = node.color
    print_mode = node.print_mode

    def _trace(source):
        def on_subscribe(observer, scheduler):
            def on_next(value):
                if trace_next is True:
                    if trace_next_payload is True:
                        print_info(
                            node_name,
                            color,
                            id,
                            "on_next",
                            value,
                            date=date or datetime.datetime.now(),
                            print_mode=print_mode,
                            log_level=DEBUG,
                        )
                    else:
                        print_info(
                            node_name,
                            color,
                            id,
                            "on_next",
                            "",
                            date=date or datetime.datetime.now(),
                            print_mode=print_mode,
                            log_level=DEBUG,
                        )
                observer.on_next(value)

            def on_completed():
                value = ""
                print_info(
                    node_name,
                    color,
                    id,
                    "on_completed",
                    value,
                    date=date or datetime.datetime.now(),
                    print_mode=print_mode,
                    log_level=DEBUG,
                )
                observer.on_completed()

            def on_error(error):
                if isinstance(error, Exception):
                    error_traceback = "%s, %s" % (
                        error,
                        traceback.print_tb(error.__traceback__),
                    )
                    print_info(
                        node_name,
                        color,
                        id,
                        "on_error",
                        error_traceback,
                        date=date or datetime.datetime.now(),
                    )
                else:
                    print_info(
                        node_name,
                        color,
                        id,
                        "on_error",
                        error,
                        date=date or datetime.datetime.now(),
                        print_mode=print_mode,
                        log_level=rospy.ERROR,
                    )
                observer.on_error(error)

            def dispose():
                if trace_subscribe is True:
                    value = ""
                    print_info(
                        node_name,
                        color,
                        id,
                        "dispose",
                        value,
                        date=date or datetime.datetime.now(),
                        print_mode=print_mode,
                        log_level=DEBUG,
                    )
                disposable.dispose()

            if trace_subscribe is True:
                value = ""
                print_info(
                    node_name,
                    color,
                    id,
                    "on_subscribe",
                    value,
                    date=date or datetime.datetime.now(),
                    print_mode=print_mode,
                    log_level=DEBUG,
                )
            disposable = source.subscribe(
                on_next=on_next,
                on_error=on_error,
                on_completed=on_completed,
            )
            return Disposable(dispose)

        return rx.create(on_subscribe)

    return _trace


def flag_dict(name):
    def _init_flag_dict(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                flag_dict = {name: value.data}
                observer.on_next(flag_dict)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _init_flag_dict


def filter_dict():
    def _filter_dict(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                d = dict()
                for node_name, flag in value.items():
                    if flag is True:
                        continue
                    d[node_name] = flag
                observer.on_next(d)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _filter_dict


def switch_to_reset():
    def _switch_to_reset(source):
        def subscribe(observer, scheduler=None):
            reset_mode = [False]

            def on_next(value):
                if isinstance(value, bool):
                    reset_mode[0] = True  # if we receive a reset flag, turn on reset mode
                    observer.on_next(value)
                else:
                    if not reset_mode[0]:  # Check if we haven't previously received a reset message
                        observer.on_next(value)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _switch_to_reset


def combine_dict(acc, x):
    for key, item in acc.items():
        item += x[key]
    return acc


def create_msg_tuple(name: str, node_tick: int, msg: List[Any], stamp: List[Stamp], done: bool = None):
    info = Info(name=name, node_tick=node_tick, t_in=stamp, done=done)
    return Msg(info, msg)


def remap_state(name, is_reactive, real_time_factor):
    def _remap_state(source):
        def subscribe(observer, scheduler=None):
            start = time.time()
            seq = [0]

            def on_next(value):
                node_tick = value[0][0]
                msg = value[0][1]
                done = value[1]
                wc_stamp = time.time()
                if is_reactive:
                    sim_stamp = None
                else:
                    sim_stamp = (wc_stamp - start) / real_time_factor
                stamp = Stamp(seq[0], sim_stamp, wc_stamp)
                res = create_msg_tuple(name, node_tick, [msg], [stamp], done=done)
                seq[0] += 1
                observer.on_next(res)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _remap_state


def remap_target(name, is_reactive, real_time_factor):
    def _remap_target(source):
        def subscribe(observer, scheduler=None):
            start = time.time()
            seq = [0]

            def on_next(value):
                node_tick = value[0]
                msg = value[1]
                wc_stamp = time.time()
                if is_reactive:
                    sim_stamp = None
                else:
                    sim_stamp = (wc_stamp - start) / real_time_factor
                stamp = Stamp(seq[0], sim_stamp, wc_stamp)
                res = create_msg_tuple(name, node_tick, [msg], [stamp])
                seq[0] += 1
                observer.on_next(res)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _remap_target


def filter_info_for_printing(info):
    info_dict = dict()
    if info.rate_in:
        info_dict["rate_in"] = info.rate_in
    info_dict["t_in"] = [t.sim_stamp for t in info.t_in]
    if info.t_node:
        info_dict["t_node"] = [t.sim_stamp for t in info.t_node]
    if info.done:
        info_dict["done"] = info.done
    return info_dict


def remap_cb_input(mode=0):
    def _remap_cb_input(value):
        # mode=0 (info only), mode=1 (msgs only), mode=2 (all)
        if mode == 2:
            return value
        if isinstance(value, tuple):
            mapped_value = tuple([value[0].copy(), value[1].copy()])
            for i in mapped_value:
                for key, msg in i.items():
                    if key not in ["node_tick", "t_n"]:
                        if mode == 0:
                            i[key] = filter_info_for_printing(msg.info)
                        else:
                            i[key] = msg.msgs
        else:
            mapped_value = value.copy()
            for key, msg in mapped_value.items():
                if key not in ["node_tick", "t_n"]:
                    if mode == 0:
                        mapped_value[key] = filter_info_for_printing(msg.info)
                    else:
                        mapped_value[key] = msg.msgs
        return mapped_value

    return _remap_cb_input


def regroup_inputs(node, rate_node=1, is_input=True, perform_checks=True):
    node_name = node.ns_name
    color = node.color
    print_mode = node.print_mode

    def _regroup_inputs(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                # Regroups all inputs into a single dict
                if is_input:
                    node_tick = value[0].info.node_tick
                    t_n = round(node_tick / rate_node, 12)
                    res = dict(node_tick=node_tick, t_n=t_n)
                else:
                    res = dict()
                for msg in value:
                    res[msg.info.name] = msg

                # Perform checks
                if perform_checks and is_input:
                    node_ticks = []
                    for msg in value:
                        node_ticks.append(msg.info.node_tick)
                    if len(node_ticks) > 0:
                        if not len(set(node_ticks)) == 1:
                            print_info(
                                node_name,
                                color,
                                "regroup_inputs",
                                trace_type="",
                                value="Not all node_ticks are the same: %s" % str(value),
                                print_mode=print_mode,
                                log_level=rospy.ERROR,
                            )

                # Send regrouped input
                observer.on_next(res)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _regroup_inputs


def expected_inputs(idx_n, rate_in, rate_node, delay):
    if idx_n < 0:
        return 0
    elif idx_n == 0:
        return 1
    else:
        # N = idx_n + 1, because idx_n starts at 0
        N_t_min_1 = idx_n
        N_t = idx_n + 1
        # Note: idx_n=1 after initial tick, corresponding to T=0
        # Hence, T_t=dt_n * (idx_n-1), T_t+1=dt_n * idx_n
        sum_t_min_1 = calculate_inputs(N_t_min_1, rate_in, rate_node, delay)
        sum_t = calculate_inputs(N_t, rate_in, rate_node, delay)
        return sum_t - sum_t_min_1


def calculate_inputs(N_node, rate_in, rate_node, delay):
    N_in = max(0, int((rate_in * (N_node - 1) - rate_node * rate_in * delay) // rate_node))  # Current timestep
    # N_in = max(0, int(((rate_in * (N_node - 1) - rate_node * rate_in * delay) / rate_node)))  # Current timestep
    # N_in = max(0, int(round(((dt_n * (N_node - 1) - delay) / dt_i), 11)))  # Current timestep
    # N_in = max(0, int(((dt_n * (N_node - 1) - delay) / dt_i)))  # Current timestep
    return N_in


def generate_msgs(
    source_Nc: Observable,
    rate_node: float,
    name: str,
    rate_in: float,
    params: dict,
    is_reactive: bool,
    real_time_factor: float,
    simulate_delays: bool,
    node=None,
):
    dt_i = 1 / rate_in

    def _generate_msgs(source_msg: Observable):
        window = params["window"]
        skip = int(params["skip"])

        def subscribe(observer: typing.Observer, scheduler: Optional[typing.Scheduler] = None) -> CompositeDisposable:
            start = time.time()
            msgs_queue: List = []
            t_i_queue: List = []
            num_queue: List = []
            tick_queue: List = []
            msgs_window = deque(maxlen=window)
            t_i_window = deque(maxlen=window)
            t_n_window = deque(maxlen=window)
            lock = RLock()

            @synchronized(lock)
            def next(i):
                if len(tick_queue) > 0:
                    if not is_reactive or len(msgs_queue) >= num_queue[0]:
                        try:
                            tick = tick_queue.pop(0)
                            if is_reactive:
                                # determine num_msgs
                                num_msgs = num_queue.pop(0)
                                msgs = msgs_queue[:num_msgs]
                                t_i = t_i_queue[:num_msgs]
                                msgs_queue[:] = msgs_queue[num_msgs:]
                                t_i_queue[:] = t_i_queue[num_msgs:]
                            else:  # Empty complete buffer
                                msgs = msgs_queue.copy()
                                t_i = t_i_queue.copy()
                                msgs_queue[:] = []
                                t_i_queue[:] = []
                        except Exception as ex:  # pylint: disable=broad-except
                            observer.on_error(ex)
                            return

                        # Determine t_n stamp
                        wc_stamp = time.time()
                        seq = tick
                        if is_reactive:
                            sim_stamp = round(tick / rate_node, 12)
                        else:
                            sim_stamp = (wc_stamp - start) / real_time_factor
                        t_n = Stamp(seq, sim_stamp, wc_stamp)

                        if window > 0:
                            msgs_window.extend(msgs)
                            t_i_window.extend(t_i)
                            t_n_window.extend([t_n] * len(msgs))
                            wmsgs = list(msgs_window)
                            wt_i = list(t_i_window)
                            wt_n = list(t_n_window)
                        else:
                            wmsgs = msgs
                            wt_i = t_i
                            wt_n = [t_n] * len(msgs)
                        res = Msg(Info(name, tick, rate_in, wt_n, wt_i, None), wmsgs)
                        observer.on_next(res)

            # Determine Nc logic
            def on_next_Nc(x):
                if is_reactive:
                    # Caculate expected number of message to be received
                    delay = params["delay"] if simulate_delays else 0.0
                    num_msgs = expected_inputs(x - skip, rate_in, rate_node, delay)
                    num_queue.append(num_msgs)
                tick_queue.append(x)
                next(x)

            subscriptions = []
            sad = SingleAssignmentDisposable()
            sad.disposable = source_Nc.subscribe(on_next_Nc, observer.on_error, observer.on_completed, scheduler)
            subscriptions.append(sad)

            def on_next_msg(x):
                msgs_queue.append(x[1])
                wc_stamp = time.time()
                seq = x[0]
                if is_reactive:
                    sim_stamp = round(x[0] * dt_i, 12)
                else:
                    sim_stamp = (wc_stamp - start) / real_time_factor
                t_i_queue.append(Stamp(seq, sim_stamp, wc_stamp))
                next(x)

            sad = SingleAssignmentDisposable()
            if not is_reactive and simulate_delays:
                source_msg_delayed = source_msg.pipe(ops.delay(params["delay"] / real_time_factor))
            else:
                source_msg_delayed = source_msg
            sad.disposable = source_msg_delayed.subscribe(on_next_msg, observer.on_error, observer.on_completed, scheduler)
            subscriptions.append(sad)

            return CompositeDisposable(subscriptions)

        return rx.create(subscribe)

    return _generate_msgs


def create_channel(
    ns,
    Nc,
    rate_node,
    inpt,
    is_reactive,
    real_time_factor,
    simulate_delays,
    E,
    scheduler,
    is_feedthrough,
    node,
):
    if is_feedthrough:
        name = inpt["feedthrough_to"]
    else:
        name = inpt["name"]

    # Readable format
    Is = inpt["reset"]
    Ir = inpt["msg"].pipe(
        ops.observe_on(scheduler),
        ops.map(inpt["converter"].convert),
        ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
        ops.share(),
    )

    # Get rate from rosparam server
    if inpt["external_rate"] and inpt["external_rate"] > 0:
        rate = inpt["external_rate"]
    else:
        rate_str = "%s/rate/%s" % (ns, inpt["address"][len(ns) + 1 :])
        rate = get_param_with_blocking(rate_str)

    # Create input channel
    if real_time_factor == 0:
        Nc = Nc.pipe(ops.observe_on(scheduler), ops.start_with(0))
    else:
        Nc = Nc.pipe(ops.observe_on(scheduler))

    channel = Ir.pipe(
        generate_msgs(
            Nc,
            rate_node,
            name,
            rate,
            params=inpt,
            is_reactive=is_reactive,
            real_time_factor=real_time_factor,
            simulate_delays=simulate_delays,
            node=node,
        ),
        ops.share(),
    )

    # Create reset flag
    flag = Ir.pipe(
        ops.map(lambda val: val[0] + 1),
        ops.start_with(0),
        ops.combine_latest(Is.pipe(ops.map(lambda msg: msg.data))),  # Depends on ROS reset msg type
        ops.filter(lambda value: not is_reactive or value[0] == value[1]),
        ops.map(lambda x: {name: x[0]}),
    )
    return channel, flag


def init_channels(
    ns,
    Nc,
    rate_node,
    inputs,
    is_reactive,
    real_time_factor,
    simulate_delays,
    E,
    scheduler,
    node,
    is_feedthrough=False,
):
    # Create channels
    channels = []
    flags = []
    for i in inputs:
        channel, flag = create_channel(
            ns,
            Nc,
            rate_node,
            i,
            is_reactive,
            real_time_factor,
            simulate_delays,
            E,
            scheduler,
            is_feedthrough,
            node,
        )
        channels.append(channel)
        if is_feedthrough:
            name = i["address"]
        else:
            name = i["name"]
        flag_name = "flag [%s]" % name.split("/")[-1][:12].ljust(4)
        # flag.pipe(spy(f'sub_{flag_name}', node)).subscribe(print)
        flag = flag.pipe(spy(flag_name, node))  # , ops.take(1), ops.merge(rx.never()))
        flags.append(flag)
    zipped_flags = rx.zip(*flags).pipe(ops.map(lambda x: merge_dicts({}, x)))
    zipped_channels = rx.zip(*channels).pipe(
        ops.combine_latest(
            E.pipe(ops.observe_on(scheduler))
        ),  # Latch output on '/end_reset' --> Can only receive 1 each episode.
        ops.map(lambda x: x[0]),
        regroup_inputs(node, rate_node=rate_node),
        ops.share(),
    )
    return zipped_channels, zipped_flags


def init_real_reset(
    ns,
    Nc,
    rate_node,
    RR,
    real_reset,
    feedthrough,
    targets,
    is_reactive,
    real_time_factor,
    simulate_delays,
    E,
    scheduler,
    node,
):
    # Create real reset pipeline
    dispose = []
    if real_reset:
        for i in feedthrough:
            rate_str = "%s/rate/%s" % (ns, i["address"][len(ns) + 1 :])
            rate_in = get_param_with_blocking(rate_str)
            if not rate_in == rate_node:
                raise ValueError(
                    "Rate of the switch node (%s) must be exactly the same as the feedthrough node rate (%s)."
                    % (rate_node, rate_node)
                )

        # Create zipped action channel
        zipped_channels, zipped_flags = init_channels(
            ns,
            Nc,
            rate_node,
            feedthrough,
            is_reactive,
            real_time_factor,
            simulate_delays,
            E,
            scheduler,
            node,
            is_feedthrough=True,
        )

        # Create switch subject
        target_signal = rx.zip(*[t["msg"] for t in targets])
        RR_ho = BehaviorSubject(zipped_channels)
        d_RR_ho = RR.pipe(
            ops.combine_latest(target_signal),  # make switch logic wait for all targets to be received.
            ops.map(lambda x: Nc.pipe(ops.map(lambda x: None), ops.start_with(None))),
        ).subscribe(RR_ho)
        rr_channel = RR_ho.pipe(ops.switch_latest())

        # Add disposables
        dispose += [RR_ho, d_RR_ho]
    else:
        # Create switch Subject
        zipped_flags = rx.never().pipe(ops.start_with({}))
        rr_channel = Nc.pipe(ops.map(lambda x: None), ops.start_with(None))

    return rr_channel, zipped_flags, dispose


def init_target_channel(states, scheduler, node):
    channels = []
    for s in states:
        c = s["msg"].pipe(
            ops.map(s["converter"].convert),
            ops.share(),
            ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
            remap_target(s["name"], node.is_reactive, node.real_time_factor),
        )
        channels.append(c)
    return rx.zip(*channels).pipe(regroup_inputs(node, is_input=False))


def merge_dicts(dict_1, dict_2):
    if isinstance(dict_2, dict):
        dict_2 = (dict_2,)
    for d in dict_2:
        dict_1.update(d)
    return dict_1


def init_state_inputs_channel(ns, state_inputs, scheduler, node):
    if len(state_inputs) > 0:
        channels = []
        for s in state_inputs:
            d = s["done"].pipe(
                ops.map(lambda msg: bool(msg.data)),
                ops.scan(lambda acc, x: x if x else acc, False),
            )
            c = s["msg"].pipe(
                ops.map(s["converter"].convert),
                ops.share(),
                ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
                ops.start_with((-1, None)),
                ops.combine_latest(d),
                ops.filter(lambda x: x[0][0] >= 0 or x[1]),
                remap_state(s["name"], node.is_reactive, node.real_time_factor),
            )
            channels.append(c)
        return rx.zip(*channels).pipe(regroup_inputs(node, is_input=False), ops.merge(rx.never()))
    else:
        return rx.never().pipe(ops.start_with(dict()))


def call_state_reset(state):
    def _call_state_reset(source):
        def subscribe(observer, scheduler=None):
            def on_next(state_msg):
                state.reset(state=state_msg.msgs[0], done=state_msg.info.done)
                observer.on_next(state_msg)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _call_state_reset


def init_state_resets(ns, state_inputs, trigger, scheduler, node):
    if len(state_inputs) > 0:
        channels = []

        for s in state_inputs:
            d = s["done"].pipe(
                ops.map(lambda msg: bool(msg.data)),
                ops.scan(lambda acc, x: x if x else acc, False),
            )
            c = s["msg"].pipe(
                ops.map(s["converter"].convert),
                ops.share(),
                ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
                ops.start_with((-1, None)),
                ops.combine_latest(d),
                ops.filter(lambda x: x[0][0] >= 0 or x[1]),
                remap_state(s["name"], node.is_reactive, node.real_time_factor),
            )

            done, reset = trigger.pipe(
                with_latest_from(c),
                ops.take(1),
                ops.merge(rx.never()),
                ops.map(lambda x: x[1]),
                ops.partition(lambda x: x.info.done),
            )
            reset = reset.pipe(call_state_reset(s["state"]))
            rs = rx.merge(
                done.pipe(spy("done [%s]" % s["name"].split("/")[-1][:12].ljust(4), node)),
                reset.pipe(spy("reset [%s]" % s["name"].split("/")[-1][:12].ljust(4), node)),
            )

            channels.append(rs)
        return rx.zip(*channels).pipe(regroup_inputs(node, is_input=False), ops.merge(rx.never()))
    else:
        return rx.never().pipe(ops.start_with(dict()))


def init_callback_pipeline(
    ns,
    cb_tick,
    cb_ft,
    stream,
    real_reset,
    targets,
    state_outputs,
    outputs,
    scheduler,
    node,
):
    d_msg = []
    if real_reset:
        target_stream = init_target_channel(targets, scheduler, node)

        # Split stream into feedthrough (ft) and reset stream
        reset_stream, ft_stream = stream.pipe(ops.partition(lambda x: x[1][1] is None))

        # Either feedthrough action or run callback
        ft_stream = ft_stream.pipe(
            ops.map(lambda x: x[1][1]),
            spy("CB_FT", node, log_level=DEBUG, mapper=remap_cb_input(mode=0)),
            ops.map(lambda val: cb_ft(val, node.is_reactive)),
            ops.share(),
        )
        reset_stream = reset_stream.pipe(
            ops.map(lambda x: x[1][0]),
            ops.combine_latest(target_stream),
            spy("CB_RESET", node, log_level=DEBUG, mapper=remap_cb_input(mode=0)),
            ops.map(lambda val: cb_tick(**val[0], **val[1])),
            ops.share(),
        )
        output_stream = rx.merge(reset_stream, ft_stream)

        # Send done flags
        for s in state_outputs:
            d = reset_stream.pipe(
                spy("reset", node, log_level=DEBUG),
                ops.pluck(s["name"] + "/done"),
                ops.share(),
            ).subscribe(s["msg"])

            # Add disposable
            d_msg += [d]
    else:
        output_stream = stream.pipe(
            ops.filter(lambda x: x[1][1] is None),
            ops.map(lambda x: x[1][0]),
            spy("CB_TICK", node, log_level=DEBUG, mapper=remap_cb_input(mode=0)),
            ops.map(lambda val: cb_tick(**val)),
            ops.share(),
        )
    return d_msg, output_stream


def get_node_params(msg):
    node_name = msg.data
    node_params = get_param_with_blocking(node_name)
    if node_params is None:
        rospy.logwarn(
            "Parameters for object registry request (%s) not found on parameter server. Timeout: object (%s) not registered."
            % (msg.data, msg.data)
        )
    return node_params


def extract_node_reset(ns, node_params, sp_nodes, launch_nodes):
    name = node_params["name"]
    nf = dict(name=name, address="%s/%s/end_reset" % (ns, name), msg=Subject(), msg_type=Bool)
    return dict(
        inputs=[],
        reactive_proxy=[],
        state_inputs=[],
        node_flags=[nf],
        sp_nodes=sp_nodes,
        launch_nodes=launch_nodes,
    )


def get_object_params(msg):
    obj_name = msg.data

    obj_params = get_param_with_blocking(obj_name)
    if obj_params is None:
        rospy.logwarn(
            "Parameters for object registry request (%s) not found on parameter server. Timeout: object (%s) not registered."
            % (msg.data, msg.data)
        )
        return None

    # Get state parameters from ROS param server
    state_params = obj_params["states"]

    # Get parameters from ROS param server
    node_params = []
    for node_name in obj_params["node_names"]:
        params = get_param_with_blocking(node_name)
        node_params.append(params)
    return obj_params, node_params, state_params


def extract_inputs_and_reactive_proxy(ns, node_params, state_params, sp_nodes, launch_nodes):
    inputs = []
    state_inputs = []
    reactive_proxy = []
    node_flags = []

    # Process states
    for i in state_params:
        name = i["name"]

        # Convert to classes
        i["msg_type"] = get_attribute_from_module(i["msg_type"])
        if "converter" in i and isinstance(i["converter"], dict):
            i["converter"] = initialize_converter(i["converter"])
        elif "converter" not in i:
            i["converter"] = Identity()

        # Initialize rx objects
        i["msg"] = Subject()  # S
        i["done"] = Subject()  # D

        # Create a new state
        s = dict()
        s.update(i)
        state_inputs.append(s)

    # Process nodes
    converted_outputs = dict()  # {address: (msg_type_out, converter, msg_type_ros, source)}
    for params in node_params:
        name = params["name"]

        # Process node flags
        nf = dict(
            name=name,
            address="%s/%s/end_reset" % (ns, name),
            msg=Subject(),
            msg_type=Bool,
        )
        node_flags.append(nf)

        for i in params["outputs"]:
            ros_msg_type = get_opposite_msg_cls(i["msg_type"], i["converter"])

            # Infer input (ROS) message type from  output msg_type and converter
            if not i["converter"] == Identity().get_yaml_definition():
                if params["name"].split("/")[-2] == "sensors":  # if output is also the sensor output
                    split_name = params["name"].split("/")
                    source = ("/".join(split_name[:-2]), split_name[-2], split_name[-1])
                else:
                    source = (params["name"], "outputs", i["name"])
                converted_outputs[i["address"]] = (
                    get_attribute_from_module(i["msg_type"]),
                    i["converter"],
                    ros_msg_type,
                    source,
                )

            # Create a new input topic for each SimNode output topic
            n = RxInput(
                name=i["address"],
                address=i["address"],
                msg_type=get_module_type_string(ros_msg_type),
                external_rate=None,
                window=0,
            ).build()

            # Convert to classes
            n["msg_type"] = get_attribute_from_module(n["msg_type"])
            if isinstance(i["converter"], dict):
                n["converter"] = initialize_converter(i["converter"])
            else:
                n["converter"] = i["converter"]

            # Initialize rx objects
            n["msg"] = Subject()  # Ir
            n["reset"] = Subject()  # Is
            inputs.append(n)

        for i in params["inputs"]:
            if i["external_rate"]:
                # Convert to classes
                i["msg_type"] = get_attribute_from_module(i["msg_type"])
                if isinstance(i["converter"], dict):
                    i["converter"] = initialize_converter(i["converter"])

                # Initialize rx reset output for reactive input
                i["reset"] = Subject()

                # Create a new output topic for each SimNode reactive input (bridge sends reset msg)
                o = dict()
                o.update(i)
                reactive_proxy.append(o)

    # Check that converted outputs do not break the object's simulation graph (some nodes might expect a non-converted output message).
    for params in node_params:
        for i in params["inputs"]:
            if i["address"] in converted_outputs:
                # determine message conversion
                msg_type_out = converted_outputs[i["address"]][0]
                converter_out = converted_outputs[i["address"]][1]
                msg_type_ros = converted_outputs[i["address"]][2]
                converter_in = i["converter"]
                msg_type_in = get_opposite_msg_cls(msg_type_ros, i["converter"])
                msg_type_in_yaml = get_attribute_from_module(i["msg_type"])

                target = (params["name"], "inputs", params["inputs"][0]["name"])
                source = converted_outputs[i["address"]][3]

                msg_type_str = msg_type_error(
                    source,
                    target,
                    msg_type_out,
                    converter_out,
                    msg_type_ros,
                    converter_in,
                    msg_type_in,
                    msg_type_in_yaml,
                )
                assert msg_type_in == msg_type_in_yaml, msg_type_str

    return dict(
        inputs=inputs,
        reactive_proxy=reactive_proxy,
        state_inputs=state_inputs,
        node_flags=node_flags,
        sp_nodes=sp_nodes,
        launch_nodes=launch_nodes,
    )


def initialize_reactive_proxy_reset(rate_node, RM, reactive_proxy, node):
    for rp in reactive_proxy:
        if "disposable" in rp:
            continue
        rate_in = rp["external_rate"]
        rp["disposable"] = RM.pipe(
            ops.map(lambda msg: msg.data),
            ops.map(
                lambda idx_n: 1 + int((idx_n - 1) * rate_in // rate_node)
            ),  # We subtract -1 from msg to account for the initial tick.
            ops.map(lambda i: UInt64(data=i)),
        ).subscribe(rp["reset"])
    return None


def switch_with_check_pipeline(init_ho=None):
    if init_ho is None:
        stream_ho = Subject()
    else:
        stream_ho = BehaviorSubject(init_ho)
    check, stream = stream_ho.pipe(ops.switch_latest(), ops.partition(lambda event: event is None))
    return check, stream, stream_ho


def node_reset_flags(ns, node_flags, node):
    flags = [
        nf["msg"].pipe(
            flag_dict(nf["name"]),
            spy("has_reset", node),
            ops.start_with({nf["name"]: False}),
        )
        for nf in node_flags
    ]
    init_dict = dict()
    for nf in node_flags:
        init_dict[nf["name"]] = False
    stream = rx.merge(*flags).pipe(
        ops.scan(lambda acc, x: merge_dicts(acc, x), init_dict),
        ops.filter(lambda x: any([value for key, value in x.items()])),
        filter_dict(),
        spy("awaiting", node, log_level=DEBUG),
        ops.filter(lambda x: len(x) == 0),
        ops.start_with(None),
    )
    return stream


def filter_dict_on_key(key):
    def _filter_dict_on_key(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                if key in value:
                    observer.on_next(value[key])

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _filter_dict_on_key


def throttle_with_time(dt, node, rate_tol: float = 0.95, log_level: int = INFO):
    time_fn = time.perf_counter
    node_name = node.ns_name
    color = node.color
    print_mode = node.print_mode
    effective_log_level = logging.getLogger("rosout").getEffectiveLevel()
    log_time = 2  # [s]

    def _throttle_with_time(source):
        def subscribe(observer, scheduler=None):
            # Timing
            tic = [None]
            cum_cbs = [0]
            cum_delay = [0]
            cum_sleep = [0]

            # Logging
            last_cum_cbs = [0]
            last_cum_delay = [0]
            last_cum_sleep = [0]
            last_Nc = [0]
            last_time = [None]

            def on_next(Nc):
                if tic[0] is None:
                    tic[0] = time_fn()
                elif last_time[0] is None:
                    last_time[0] = time_fn()
                toc = time_fn()
                sleep_time = dt - (toc - tic[0])
                if sleep_time > 0:  # sleep if overdue is negative
                    time.sleep(sleep_time)
                    cum_sleep[0] += sleep_time
                else:  # If we are overdue, then next tick is shifted by overdue
                    cum_delay[0] += -sleep_time
                    cum_cbs[0] += 1
                tic[0] = toc + max(sleep_time, 0)

                # Logging
                curr = time_fn()
                if last_time[0] and (curr - last_time[0]) > log_time:
                    # Calculate statistics since last logged instance
                    log_window = curr - last_time[0]
                    log_cbs = cum_cbs[0] - last_cum_cbs[0]
                    # log_delay = cum_delay[0] - last_cum_delay[0]
                    log_sleep = cum_sleep[0] - last_cum_sleep[0]
                    log_Nc = Nc - last_Nc[0]

                    # Log statistics if not keeping rate
                    Nc_expected = log_window / dt  # [ticks]
                    rate_ratio = log_Nc / Nc_expected
                    cbs_ratio = log_cbs / log_Nc
                    sleep_ratio = log_sleep / log_window
                    # delay_ratio = log_delay / log_window
                    if rate_ratio < rate_tol and node.log_level >= effective_log_level and WARN >= effective_log_level:
                        print_str = f"Running at {rate_ratio*100:.2f}% of rate ({1/dt} Hz) | {sleep_ratio*100:.2f}% sleep | {100 - sleep_ratio*100:.2f}% computation | {cbs_ratio*100: .2f}% callbacks delayed |"
                        print_info(
                            node_name,
                            "red",
                            f"last {log_window:.2f} s",
                            trace_type="",
                            value=print_str,
                            print_mode=print_mode,
                            log_level=WARN,
                        )
                    elif node.log_level >= effective_log_level and log_level >= effective_log_level:
                        print_str = f"Running at {rate_ratio*100:.2f}% of rate ({1/dt} Hz) | {sleep_ratio*100:.2f}% sleep | {100 - sleep_ratio*100:.2f}% computation | {cbs_ratio*100: .2f}% callbacks delayed |"
                        print_info(
                            node_name,
                            color,
                            f"last {log_window:.2f} s",
                            trace_type="",
                            value=print_str,
                            print_mode=print_mode,
                            log_level=log_level,
                        )

                    # Set baseline statistics
                    last_time[0] = curr
                    last_cum_cbs[0] = cum_cbs[0]
                    last_cum_delay[0] = cum_delay[0]
                    last_cum_sleep[0] = cum_sleep[0]
                    last_Nc[0] = Nc

                # Send tick for next callback
                observer.on_next(Nc)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _throttle_with_time


def throttle_callback_trigger(rate_node, Nc, E, is_reactive, real_time_factor, scheduler, node):
    # return Nc
    if is_reactive and real_time_factor == 0:
        Nct = Nc
    else:
        assert (
            real_time_factor > 0
        ), "The real_time_factor must be larger than zero when *not* running reactive (i.e. asychronous)."
        wc_dt = 1 / (rate_node * real_time_factor)
        Nct = Nc.pipe(
            ops.scan(lambda acc, x: acc + 1, 0),
            ops.start_with(0),
            ops.observe_on(scheduler),
            throttle_with_time(wc_dt, node),
            ops.share(),
        )
    return Nct


def with_latest_from(*sources: Observable):
    def _with_latest_from(parent: Observable) -> Observable:
        NO_VALUE = NotSet()

        def subscribe(observer, scheduler=None):
            def subscribe_all(parent, *children):
                parent_queued = [None]
                values = [NO_VALUE for _ in children]

                def subscribe_child(i, child):
                    subscription = SingleAssignmentDisposable()

                    def on_next(value):
                        with parent.lock:
                            values[i] = value
                            if parent_queued[0] is not None and NO_VALUE not in values:
                                result = (parent_queued[0],) + tuple(values)
                                parent_queued[0] = None
                                observer.on_next(result)

                    subscription.disposable = child.subscribe_(on_next, observer.on_error, scheduler=scheduler)
                    return subscription

                parent_subscription = SingleAssignmentDisposable()

                def on_next(value):
                    with parent.lock:
                        if NO_VALUE not in values:
                            result = (value,) + tuple(values)
                            observer.on_next(result)
                        else:
                            parent_queued[0] = value

                disp = parent.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
                parent_subscription.disposable = disp

                children_subscription = [subscribe_child(i, child) for i, child in enumerate(children)]

                return [parent_subscription] + children_subscription

            return CompositeDisposable(subscribe_all(parent, *sources))

        return Observable(subscribe)

    return _with_latest_from


class NotSet:
    """Sentinel value."""

    def __eq__(self, other):
        return self is other

    def __repr__(self):
        return "NotSet"



import eagerx.bridges.openai_gym.bridge  # noqa: F401
import eagerx.bridges.openai_gym.converters  # noqa: F401
import eagerx.bridges.openai_gym.enginenodes  # noqa: F401
import eagerx.bridges.openai_gym.objects  # noqa: F401
import eagerx.bridges.openai_gym.simstates  # noqa: F401
from eagerx.bridges.openai_gym.env import EagerGym  # noqa: F401



# ROS IMPORTS
from std_msgs.msg import UInt64, Bool, String

# RX IMPORTS
import rx
from rx import operators as ops
from rx.disposable import CompositeDisposable
from rx.scheduler import EventLoopScheduler, ThreadPoolScheduler
from rx.subject import ReplaySubject, Subject, BehaviorSubject

# EAGERX IMPORTS
from eagerx.core.constants import DEBUG

from eagerx.core.rx_operators import (
    cb_ft,
    spy,
    trace_observable,
    flag_dict,
    switch_to_reset,
    combine_dict,
    init_channels,
    init_real_reset,
    merge_dicts,
    init_state_inputs_channel,
    init_state_resets,
    init_callback_pipeline,
    get_object_params,
    extract_inputs_and_reactive_proxy,
    initialize_reactive_proxy_reset,
    switch_with_check_pipeline,
    node_reset_flags,
    filter_dict_on_key,
    get_node_params,
    extract_node_reset,
    throttle_callback_trigger,
    with_latest_from,
)


def init_node_pipeline(
    ns,
    rate_node,
    node,
    inputs,
    outputs,
    F,
    SS_ho,
    SS_CL_ho,
    R,
    RR,
    E,
    real_reset,
    feedthrough,
    state_inputs,
    state_outputs,
    targets,
    cb_ft,
    is_reactive,
    real_time_factor,
    simulate_delays,
    disposables,
    event_scheduler=None,
):
    # Node ticks
    Rn = ReplaySubject()  # Reset flag for the node (Nc=Ns and r_signal)
    Nc = Subject()  # Number completed callbacks (i.e. send Topics):
    Ns = BehaviorSubject(0)  # Number of started callbacks (i.e. number of planned Topics).

    # Throttle the callback trigger
    Nct = throttle_callback_trigger(rate_node, Nc, E, is_reactive, real_time_factor, event_scheduler, node)

    # Create input channels
    zipped_inputs, zipped_input_flags = init_channels(
        ns,
        Nct,
        rate_node,
        inputs,
        is_reactive,
        real_time_factor,
        simulate_delays,
        E,
        node=node,
        scheduler=event_scheduler,
    )

    # Create action channels
    zipped_feedthrough, zipped_action_flags, d_rr = init_real_reset(
        ns,
        Nct,
        rate_node,
        RR,
        real_reset,
        feedthrough,
        targets,
        is_reactive,
        real_time_factor,
        simulate_delays,
        E,
        event_scheduler,
        node=node,
    )

    # Zip inputs & action channels
    zipped_channels = rx.zip(zipped_inputs, zipped_feedthrough).pipe(
        ops.share(), ops.observe_on(event_scheduler)
    )  # this is required, otherwise a block.

    # New routine with merge
    PR = R.pipe(
        ops.observe_on(event_scheduler),
        ops.map(lambda x: True),
        ops.merge(zipped_channels),
        switch_to_reset(),
        ops.share(),
    )

    # Create reset signal
    Rr, P = PR.pipe(ops.partition(lambda value: isinstance(value, bool)))

    # Create accumulator: (acc)
    d_Ns = P.pipe(ops.scan(lambda acc, x: acc + 1, 0)).subscribe(Ns)

    # Create callback stream
    input_stream = Ns.pipe(ops.skip(1), ops.zip(P), ops.share())
    d_msg, output_stream = init_callback_pipeline(
        ns,
        node.callback_cb,
        cb_ft,
        input_stream,
        real_reset,
        targets,
        state_outputs,
        outputs,
        event_scheduler,
        node=node,
    )

    # Publish output msg as ROS topic and to subjects if single process
    for o in outputs:
        d = output_stream.pipe(
            ops.filter(lambda x: x is not None),
            ops.pluck(o["name"]),
            ops.filter(lambda x: x is not None),
            ops.map(o["converter"].convert),
            ops.share(),
        ).subscribe(o["msg"])
        # Add disposable
        d_msg += [d]

    # Publish output msg as ROS topic and to subjects if single process
    Nc_obs = output_stream.pipe(ops.scan(lambda acc, x: acc + 1, 0))
    Nc_empty = output_stream.pipe(ops.scan(lambda acc, x: acc + 1 if x is None else acc, 0), ops.start_with(0))

    # Increase ticks
    d_Nc = Nc_obs.subscribe(Nc, scheduler=event_scheduler)
    d_Rn = Nc_obs.pipe(
        ops.start_with(0),  # added to simulated first zero from BS(0) of Nc
        ops.zip(Nc_empty),
        ops.map(lambda x: x[0] - x[1]),
        ops.combine_latest(Ns, Rr, Nc_empty),
        ops.filter(lambda value: value[0] == value[1] - value[3]),
        ops.take(1),
        ops.merge(rx.never()),
        # spy('post-filter', node),
    ).subscribe(Rn)

    # Create reset flags for the set_states
    ss_flags = init_state_inputs_channel(ns, state_inputs, event_scheduler, node=node)
    SS_ho.on_next(ss_flags.pipe(ops.take(1), ops.start_with(None)))
    SS_CL_ho.on_next(ss_flags.pipe(ops.start_with(None)))

    # Combine flags and subscribe F to all the flags zipped together
    d_flag = (
        rx.zip(zipped_input_flags, zipped_action_flags)
        .pipe(ops.map(lambda x: merge_dicts(x[0], x[1])))
        .subscribe(F, scheduler=event_scheduler)
    )

    # Dispose
    disposables.clear()
    d = CompositeDisposable([Rn, Nc, Ns, d_Rn, d_Nc, d_Ns, d_flag] + d_msg + d_rr)
    disposables.add(d)
    return {"Rn": Rn, "dispose": d}


def init_node(
    ns,
    rate_node,
    node,
    inputs,
    outputs,
    feedthrough=tuple(),
    state_inputs=tuple(),
    targets=tuple(),
):
    # Initialize scheduler
    event_scheduler = EventLoopScheduler()
    eps_disp = CompositeDisposable()
    reset_disp = CompositeDisposable(eps_disp)

    # Gather reactive properties
    is_reactive = node.is_reactive
    real_time_factor = node.real_time_factor
    simulate_delays = node.simulate_delays

    # Track node I/O
    node_inputs = []
    node_outputs = []

    # Prepare reset topic
    R = Subject()
    reset_input = dict(name="reset", address=ns + "/reset", msg_type=UInt64, msg=R)
    node_inputs.append(reset_input)

    # Prepare real_reset topic
    RR = Subject()
    real_reset_input = dict(name="real_reset", address=ns + "/real_reset", msg_type=UInt64, msg=RR)
    node_inputs.append(real_reset_input)

    # End reset
    E = Subject()
    end_reset = dict(name="end_reset", address=ns + "/end_reset", msg_type=UInt64, msg=E)
    node_inputs.append(end_reset)

    # Prepare node reset topic
    node_reset = dict(name=node.ns_name, msg_type=Bool)
    node_reset["address"] = node.ns_name + "/end_reset"
    node_reset["msg"] = Subject()
    node_outputs.append(node_reset)

    # Real reset checks
    state_outputs = []
    real_reset = len(feedthrough) > 0
    assert not real_reset or (
        real_reset and len(state_inputs) > 0
    ), "Cannot initialize real reset node (%s). If len(feedthrough) is provided, then len(states) > 0. must hold."

    # Prepare input topics
    flags = []
    for i in inputs:
        # Subscribe to input topic
        Ir = Subject()
        i["msg"] = Ir

        # Subscribe to input reset topic
        Is = Subject()
        i["reset"] = Is

        # Prepare initial flag
        flag = i["reset"].pipe(flag_dict(i["name"]), ops.first(), ops.merge(rx.never()))
        flags.append(flag)

    # Prepare output topics
    for i in outputs:
        # Prepare output topic
        i["msg"] = Subject()

        # Initialize reset topic
        i["reset"] = Subject()

    # Prepare action topics (used by RealResetNode)
    for i in feedthrough:
        # Subscribe to input topic
        Ar = Subject()
        i["msg"] = Ar

        # Subscribe to input reset topic
        As = Subject()
        i["reset"] = As

        # Prepare initial flag
        flag = i["reset"].pipe(flag_dict(i["address"]), ops.first(), ops.merge(rx.never()))
        flags.append(flag)

    # Prepare state topics
    for i in state_inputs:
        # Initialize desired state message
        S = Subject()
        i["msg"] = S

        D = Subject()
        i["done"] = D

    for i in targets:
        # Initialize target state message
        S = Subject()
        i["msg"] = S

        D = Subject()
        i["done"] = D

        # Initialize done flag for desired state
        done_output = dict(name=i["name"], address=i["address"] + "/done", msg_type=Bool, msg=D)
        state_outputs.append(done_output)

    # Prepare initial flags
    F_init = rx.zip(*flags).pipe(spy("zip", node), ops.map(lambda x: merge_dicts({}, x)))

    # Reset flags
    F = Subject()
    f = rx.merge(F, F_init)
    ss_flags = init_state_inputs_channel(ns, state_inputs, event_scheduler, node=node)
    check_SS_CL, SS_CL, SS_CL_ho = switch_with_check_pipeline(init_ho=ss_flags)
    latched_ss_flags = rx.combine_latest(SS_CL, ss_flags).pipe(ops.map(lambda x: x[1]), ops.take(1))
    check_SS, SS, SS_ho = switch_with_check_pipeline(init_ho=latched_ss_flags)

    # Node ticks
    Rn_ho = BehaviorSubject(BehaviorSubject((0, 0, True)))
    Rn = Rn_ho.pipe(ops.switch_latest(), ops.map(lambda x: x[0]))

    # Create reset switch_latest
    Rr = R.pipe(ops.map(lambda x: True))  # ops.observe_on(event_scheduler),  # seemed to make ROS variant crash

    # Prepare "ready-to-reset" signal (i.e. after node receives reset signal and stops sending any output msgs).
    RrRn = rx.zip(Rr.pipe(spy("Rr", node)), Rn.pipe(spy("Rn", node))).pipe(
        ops.map(lambda x: x[1]), spy("SEND_FLAGS", node), ops.share()
    )

    # Send output flags
    for i in outputs:
        d = RrRn.pipe(ops.map(lambda x: UInt64(data=x))).subscribe(i["reset"])
        reset_disp.add(d)

    # Reset node pipeline
    reset_trigger = rx.zip(RrRn, f.pipe(spy("F", node)), SS.pipe(spy("SS", node))).pipe(
        with_latest_from(SS_CL),
        ops.map(lambda x: x[0][:-1] + (x[1],)),
        spy("RENEW_PIPE", node),
        ops.map(lambda x: x[-1]),
        ops.share(),
    )  # x: SS_CL
    reset_obs = reset_trigger.pipe(
        ops.map(
            lambda x: init_node_pipeline(
                ns,
                rate_node,
                node,
                inputs,
                outputs,
                F,
                SS_ho,
                SS_CL_ho,
                R,
                RR,
                E,
                real_reset,
                feedthrough,
                state_inputs,
                state_outputs,
                targets,
                cb_ft,
                is_reactive,
                real_time_factor,
                simulate_delays,
                eps_disp,
                event_scheduler=event_scheduler,
            )
        ),
        trace_observable("init_node_pipeline", node),
        ops.share(),
    )
    d = reset_obs.pipe(ops.pluck("Rn")).subscribe(Rn_ho)
    reset_disp.add(d)

    # Dispose old pipeline, run reset callback
    reset_msg = reset_obs.pipe(
        ops.pluck("dispose"),
        ops.buffer_with_count(2, skip=1),
        ops.start_with(None),
        # zipped with Rn_ho so that Rn has switched before sending "reset topic"
        ops.zip(reset_trigger, Rn_ho.pipe(ops.skip(1)), check_SS, check_SS_CL),
        ops.map(lambda x: x[1]),  # x[1]=Nc
        ops.share(),
        spy("RESET", node, log_level=DEBUG),
        ops.map(lambda x: node.reset_cb(**x)),
        trace_observable("cb_reset", node),
        ops.share(),
    )

    # Send node reset message
    d = reset_msg.pipe(ops.map(lambda x: Bool(data=True))).subscribe(node_reset["msg"])
    reset_disp.add(d)

    rx_objects = dict(
        inputs=inputs,
        outputs=outputs,
        feedthrough=feedthrough,
        state_inputs=state_inputs,
        state_outputs=state_outputs,
        targets=targets,
        node_inputs=node_inputs,
        node_outputs=node_outputs,
        disposable=reset_disp,
    )
    return rx_objects


def init_bridge_pipeline(
    ns,
    rate_node,
    node,
    zipped_channels,
    outputs,
    Nct_ho,
    DF,
    RRn_ho,
    SS_ho,
    SS_CL_ho,
    state_inputs,
    is_reactive,
    real_time_factor,
    simulate_delays,
    E,
    disposables,
    event_scheduler=None,
):
    # Node ticks
    RRn = Subject()
    RRn_ho.on_next(RRn)
    Nc = Subject()  # Number completed callbacks (i.e. send Topics): initialized at zero to kick of chain reaction
    Ns = BehaviorSubject(0)  # Number of started callbacks (i.e. number of planned Topics).

    # Throttle the callback trigger
    Nct = throttle_callback_trigger(rate_node, Nc, E, is_reactive, real_time_factor, event_scheduler, node)
    Nct_ho.on_next(Nct)

    # Create a tuple with None, to be consistent with feedthrough pipeline of init_node_pipeline
    zipped_channels = zipped_channels.pipe(ops.map(lambda i: (i, None)))

    # New routine with merge
    PR = DF.pipe(
        ops.filter(lambda x: all([df for df in x])),
        spy("DF filtered", node),
        ops.observe_on(event_scheduler),
        ops.map(lambda x: True),
        ops.merge(zipped_channels),
        switch_to_reset(),
        ops.share(),
    )

    # Create reset signal
    RRr, P = PR.pipe(ops.partition(lambda value: isinstance(value, bool)))

    # Create accumulator: (acc)
    d_Ns = P.pipe(ops.scan(lambda acc, x: acc + 1, 0)).subscribe(Ns)

    # Create callback stream
    input_stream = Ns.pipe(ops.skip(1), ops.zip(P), ops.share())
    d_msg, output_stream = init_callback_pipeline(
        ns,
        node.callback_cb,
        cb_ft,
        input_stream,
        False,
        tuple(),
        tuple(),
        outputs,
        event_scheduler,
        node,
    )

    # Publish output msg as ROS topic and to subjects if single process
    for o in outputs:
        d = output_stream.pipe(
            ops.pluck(o["name"]),
            ops.filter(lambda x: x is not None),
            ops.map(o["converter"].convert),
            ops.share(),
        ).subscribe(o["msg"])
        # Add disposable
        d_msg += [d]

    # After outputs have been send, increase the completed callback counter
    Nc_obs = output_stream.pipe(ops.scan(lambda acc, x: acc + 1, 0))

    # Increase ticks
    d_Nc = Nc_obs.subscribe(Nc, scheduler=event_scheduler)
    d_RRn = Nc_obs.pipe(
        ops.start_with(0),  # added to simulated first zero from BS(0) of Nc
        ops.combine_latest(Ns, RRr),
        ops.filter(lambda value: value[0] == value[1]),
        ops.take(1),
        ops.merge(rx.never()),
    ).subscribe(RRn)

    # Create reset flags for the set_states
    ss_flags = init_state_inputs_channel(ns, state_inputs, event_scheduler, node=node)
    SS_ho.on_next(ss_flags.pipe(ops.take(1), ops.start_with(None)))
    SS_CL_ho.on_next(ss_flags.pipe(ops.start_with(None)))

    # Dispose
    disposables.clear()
    d = CompositeDisposable([RRn, Nc, Ns, d_RRn, d_Nc, d_Ns] + d_msg)
    disposables.add(d)
    return {"dispose": d}


def init_bridge(
    ns,
    rate_node,
    node,
    inputs_init,
    outputs,
    state_inputs,
    node_names,
    target_addresses,
    message_broker,
):
    ###########################################################################
    # Initialization ##########################################################
    ###########################################################################
    # Prepare scheduler
    event_scheduler = EventLoopScheduler()
    eps_disp = CompositeDisposable()
    reset_disp = CompositeDisposable(eps_disp)

    # Gather reactive properties
    is_reactive = node.is_reactive
    real_time_factor = node.real_time_factor
    simulate_delays = node.simulate_delays

    # Prepare output topics
    for i in outputs:
        # Prepare output topic
        i["msg"] = Subject()

        # Initialize reset topic
        i["reset"] = Subject()

    # Prepare input topics
    for i in inputs_init:
        # Subscribe to input topic
        Ir = Subject()
        i["msg"] = Ir

        # Subscribe to input reset topic
        Is = Subject()
        i["reset"] = Is

    # Prepare state topics
    for i in state_inputs:
        # Initialize desired state message
        S = Subject()
        i["msg"] = S

        D = Subject()
        i["done"] = D

    ss_flags = init_state_inputs_channel(ns, state_inputs, event_scheduler, node=node)
    check_SS_CL, SS_CL, SS_CL_ho = switch_with_check_pipeline(init_ho=ss_flags)
    latched_ss_flags = rx.combine_latest(SS_CL, ss_flags).pipe(ops.map(lambda x: x[1]), ops.take(1))
    check_SS, SS, SS_ho = switch_with_check_pipeline(init_ho=latched_ss_flags)

    # Prepare target_addresses
    df_inputs = []
    dfs = []
    for i in target_addresses:
        address = "%s/%s" % (ns, i)
        done = Subject()
        df_inputs.append(dict(address=address, done=done))
        dfs.append(done.pipe(ops.map(lambda msg: bool(msg.data))))

    # Track node I/O
    node_inputs = []
    node_outputs = []

    ###########################################################################
    # Registry: node flags ####################################################
    ###########################################################################
    init_node_flags = []
    for i in node_names:
        nf = dict(name=i, address="%s/%s/end_reset" % (ns, i), msg=Subject(), msg_type=Bool)
        node_inputs.append(nf)
        init_node_flags.append(nf)

    ###########################################################################
    # Registry: real reset states #############################################
    ###########################################################################
    # Resettable real states (to dynamically add state/done flags to DF for RR)
    RR = Subject()
    RRr = RR.pipe(ops.map(lambda x: True))

    # Switch to latest state reset done flags
    DF = rx.combine_latest(RRr, *dfs)

    ###########################################################################
    # Registry ################################################################
    ###########################################################################
    # Object register (to dynamically add input reset flags to F for reset)
    NR = Subject()
    node_registry = dict(name="register_node", address=ns + "/register_node", msg=NR, msg_type=String)
    node_inputs.append(node_registry)

    # Node registry pipeline
    node_params = NR.pipe(
        spy("nodes", node, log_level=DEBUG),
        ops.map(get_node_params),
        ops.filter(lambda params: params is not None),
        ops.map(node.register_node),
        ops.map(lambda args: extract_node_reset(ns, *args)),
        ops.share(),
    )

    # Object register (to dynamically add input reset flags to F for reset)
    OR = Subject()
    object_registry = dict(name="register_object", address=ns + "/register_object", msg=OR, msg_type=String)
    node_inputs.append(object_registry)

    # Object registry pipeline
    object_params = OR.pipe(
        ops.map(get_object_params),
        ops.filter(lambda params: params is not None),
        ops.map(lambda params: (params[1],) + node.register_object(*params)),
        ops.map(lambda i: extract_inputs_and_reactive_proxy(ns, *i)),
        ops.share(),
    )

    rx_objects = rx.merge(object_params, node_params).pipe(
        ops.map(
            lambda i: (
                i,
                message_broker.add_rx_objects(
                    node.ns_name,
                    inputs=i["inputs"],
                    reactive_proxy=i["reactive_proxy"],
                    state_inputs=i["state_inputs"],
                    node_inputs=i["node_flags"],
                ),
            )
        ),
        ops.map(lambda i: i[0]),
        ops.scan(
            combine_dict,
            dict(
                inputs=inputs_init,
                reactive_proxy=[],
                sp_nodes=[],
                launch_nodes=[],
                state_inputs=[],
                node_flags=init_node_flags,
            ),
        ),
        ops.share(),
    )

    # Make sure that all nodes are initialized, before passing it to start_reset_input
    REG_cum = rx.merge(OR, NR).pipe(ops.scan(lambda acc, x: acc + 1, 0))
    rx_objects = rx.zip(rx_objects, REG_cum)

    ###########################################################################
    # Start reset #############################################################
    ###########################################################################
    # Prepare start_reset input
    SR = Subject()
    start_reset_input = dict(name="start_reset", address=ns + "/start_reset", msg=SR, msg_type=UInt64)
    node_inputs.append(start_reset_input)

    # Latch on '/rx/start_reset' event
    rx_objects = SR.pipe(
        ops.combine_latest(REG_cum),
        ops.filter(lambda x: x[0].data == x[1]),  # cum_registered == REG_cum
        ops.combine_latest(rx_objects),
        spy("SR", node, log_level=DEBUG, mapper=lambda x: (x[0][0], x[0][1], x[1][1])),
        ops.filter(lambda x: x[0][1] == x[1][1]),  # cum_registered == REG_cum && REG_cum == rx_objects[1]
        ops.map(lambda i: i[1][0]),  # rx_objects
        ops.share(),
    )
    inputs = rx_objects.pipe(ops.pluck("inputs"))
    simstate_inputs = rx_objects.pipe(ops.pluck("state_inputs"))
    reactive_proxy = rx_objects.pipe(ops.pluck("reactive_proxy"))
    node_flags = rx_objects.pipe(ops.pluck("node_flags"))

    # Prepare output for reactive proxy
    RM = Subject()
    check_reactive_proxy = reactive_proxy.pipe(ops.map(lambda rx: initialize_reactive_proxy_reset(rate_node, RM, rx, node)))

    # Zip initial input flags
    check_F_init, F_init, F_init_ho = switch_with_check_pipeline()
    F_init = F_init.pipe(ops.first(), ops.merge(rx.never()))
    d = inputs.pipe(
        ops.map(
            lambda inputs: rx.zip(*[i["reset"].pipe(flag_dict(i["name"])) for i in inputs]).pipe(
                ops.map(lambda x: merge_dicts({}, x)), ops.start_with(None)
            )
        )
    ).subscribe(F_init_ho)
    reset_disp.add(d)
    F = Subject()
    f = rx.merge(F, F_init)

    # Zip node flags
    check_NF, NF, NF_ho = switch_with_check_pipeline()
    d = node_flags.pipe(ops.map(lambda node_flags: node_reset_flags(ns, node_flags, node))).subscribe(NF_ho)
    reset_disp.add(d)

    # Dynamically initialize new state pipeline
    ResetTrigger = Subject()
    ss_flags = simstate_inputs.pipe(
        ops.map(lambda s: init_state_resets(ns, s, ResetTrigger, event_scheduler, node)),
        ops.share(),
    )
    check_simSS, simSS, simSS_ho = switch_with_check_pipeline()
    d = ss_flags.pipe(ops.map(lambda obs: obs.pipe(ops.start_with(None)))).subscribe(simSS_ho)
    reset_disp.add(d)

    # Before starting real_reset procedure, wait for EngineState pipeline to be initialized.
    # This, so that the first time, the engine states are run.
    # Some bridges/simulators might require that for setting initial state.
    ER = Subject()
    end_register = dict(name="end_register", address=ns + "/end_register", msg=ER, msg_type=UInt64)
    node_outputs.append(end_register)

    # Zip switch checks to indicate end of '/rx/start_reset' procedure, and start of '/rx/real_reset'
    d = (
        rx.zip(check_F_init, check_reactive_proxy, check_NF, check_simSS)
        .pipe(ops.map(lambda i: message_broker.connect_io()), ops.map(lambda i: UInt64()))
        .subscribe(ER)
    )
    reset_disp.add(d)

    ###########################################################################
    # Real reset ##############################################################
    ###########################################################################
    # Prepare real_reset output
    real_reset_input = dict(name="real_reset", address=ns + "/real_reset", msg=RR, msg_type=UInt64)
    node_inputs.append(real_reset_input)

    # Zip switch checks to indicate end of '/rx/start_reset' procedure, and start of '/rx/real_reset'
    d = (
        rx.zip(check_F_init, check_reactive_proxy, check_NF)
        .pipe(
            ops.map(lambda i: message_broker.connect_io()),
        )
        .subscribe(RR)
    )
    reset_disp.add(d)

    # Real reset routine. Cuts-off tick_callback when RRr is received, instead of Rr
    check_RRn, RRn, RRn_ho = switch_with_check_pipeline(init_ho=BehaviorSubject((0, 0, True)))
    pre_reset_trigger = rx.zip(RRn.pipe(spy("RRn", node)), RRr.pipe(spy("RRr", node)), SS.pipe(spy("SS", node))).pipe(
        with_latest_from(SS_CL),
        ops.map(lambda x: x[0][:-1] + (x[1],)),
        ops.map(lambda x: (x, node.pre_reset_cb(**x[-1]))),
        # Run pre-reset callback
        spy("PRE-RESET", node, log_level=DEBUG),
        trace_observable("cb_pre_reset", node),
        ops.share(),
    )

    d = pre_reset_trigger.pipe(
        ops.map(lambda x: x[0][0]),
        ops.map(lambda x: UInt64(data=x[0] + 1)),
        ops.share(),  # x[0][0]=Nc
    ).subscribe(RM)
    reset_disp.add(d)
    ss_cl = pre_reset_trigger.pipe(ops.map(lambda x: x[0][-1]))  # x[0][-1]=ss_cl

    ###########################################################################
    # Reset ###################################################################
    ###########################################################################
    # Prepare reset output
    R = Subject()
    reset_output = dict(name="reset", address=ns + "/reset", msg=R, msg_type=UInt64)
    node_outputs.append(reset_output)

    # Send reset message
    d = RM.subscribe(R)
    reset_disp.add(d)
    Rr = R.pipe(ops.map(lambda x: True))
    d = rx.zip(f.pipe(spy("F", node)), Rr.pipe(spy("Rr", node))).pipe(ops.share()).subscribe(ResetTrigger)
    reset_disp.add(d)

    # Send reset messages for all outputs (Only '/rx/bridge/outputs/tick')
    [reset_disp.add(RM.subscribe(o["reset"])) for o in outputs]

    ###########################################################################
    # Reset: initialize episode pipeline ######################################
    ###########################################################################
    # Prepare end_reset output
    end_reset = dict(name="end_reset", address=ns + "/end_reset", msg=Subject(), msg_type=UInt64)
    node_outputs.append(end_reset)

    # Dynamically initialize new input pipeline
    check_Nct, Nct, Nct_ho = switch_with_check_pipeline()
    inputs_flags = inputs.pipe(
        ops.zip(ResetTrigger),
        ops.map(lambda i: i[0]),
        ops.map(
            lambda inputs: init_channels(
                ns,
                Nct,
                rate_node,
                inputs,
                is_reactive,
                real_time_factor,
                simulate_delays,
                end_reset["msg"],
                event_scheduler,
                node,
            )
        ),
        ops.share(),
    )

    # Switch to latest zipped inputs pipeline
    check_z_inputs, z_inputs, z_inputs_ho = switch_with_check_pipeline()
    d = inputs_flags.pipe(ops.map(lambda i: i[0].pipe(ops.start_with(None)))).subscribe(z_inputs_ho, scheduler=event_scheduler)
    reset_disp.add(d)

    # Switch to latest zipped flags pipeline
    check_z_flags, z_flags, z_flags_ho = switch_with_check_pipeline()
    d = inputs_flags.pipe(ops.map(lambda i: i[1].pipe(ops.start_with(None)))).subscribe(z_flags_ho, scheduler=event_scheduler)
    reset_disp.add(d)
    d = z_flags.subscribe(F)
    reset_disp.add(d)

    # Initialize rest of episode pipeline
    pipeline_trigger = rx.zip(check_z_flags, check_z_inputs)
    reset_obs = pipeline_trigger.pipe(
        ops.map(
            lambda x: init_bridge_pipeline(
                ns,
                rate_node,
                node,
                z_inputs,
                outputs,
                Nct_ho,
                DF,
                RRn_ho,
                SS_ho,
                SS_CL_ho,
                state_inputs,
                is_reactive,
                real_time_factor,
                simulate_delays,
                end_reset["msg"],
                eps_disp,
                event_scheduler=event_scheduler,
            )
        ),
        trace_observable("init_bridge_pipeline", node),
        ops.share(),
    )

    ###########################################################################
    # End reset ###############################################################
    ###########################################################################
    # Send '/end_reset' after reset has finished
    d = reset_obs.pipe(
        ops.pluck("dispose"),
        ops.buffer_with_count(2, skip=1),
        # ops.map(lambda x: x[0].dispose()),
        ops.start_with(None),
        ops.zip(
            ss_cl,
            reset_obs,
            simSS,
            NF.pipe(spy("NF", node)),
            check_SS,
            check_SS_CL,
        ),
        ops.map(lambda x: node.reset_cb(**x[1])),
        spy("POST-RESET", node, log_level=DEBUG),
        trace_observable("cb_post_reset", node),
        ops.map(lambda x: UInt64(data=0)),
        ops.share(),
    ).subscribe(end_reset["msg"])
    reset_disp.add(d)

    rx_objects = dict(
        inputs=inputs_init,
        outputs=outputs,
        node_inputs=node_inputs,
        node_outputs=node_outputs,
        state_inputs=list(state_inputs) + df_inputs,
        disposable=reset_disp,
    )
    return rx_objects


def init_supervisor(ns, node, outputs=tuple(), state_outputs=tuple()):
    # Initialize schedulers
    tp_scheduler = ThreadPoolScheduler(max_workers=5)
    reset_disp = CompositeDisposable()

    # Prepare states
    done_outputs = []
    for s in state_outputs:
        # Prepare done flag
        s["done"] = Subject()
        done_outputs.append(
            dict(
                name=s["name"],
                address=s["address"] + "/done",
                msg_type=Bool,
                msg=s["done"],
            )
        )

        # Prepare state message (IMPORTANT: after done flag, we modify address afterwards)
        s["msg"] = Subject()
        s["address"] += "/set"

    ###########################################################################
    # Start reset #############################################################
    ###########################################################################
    SR = Subject()  # ---> Not a node output, but used in node.reset() to kickstart reset pipeline (send self.cum_registered).
    start_reset = dict(name="start_reset", address=ns + "/start_reset", msg=Subject(), msg_type=UInt64)
    d = SR.subscribe(start_reset["msg"], scheduler=tp_scheduler)
    reset_disp.add(d)

    ###########################################################################
    # End register ############################################################
    ###########################################################################
    ER = Subject()
    end_register = dict(name="reset", address=ns + "/end_register", msg_type=UInt64, msg=ER)
    real_reset = dict(name="real_reset", address=ns + "/real_reset", msg=Subject(), msg_type=UInt64)
    d = ER.subscribe(real_reset["msg"])
    reset_disp.add(d)

    # Publish state msgs
    # msgs = SR.pipe(ops.skip(1), ops.map(node._get_states), ops.share())
    msgs = ER.pipe(ops.map(node._get_states), ops.share())
    for s in state_outputs:
        d = msgs.pipe(
            ops.pluck(s["name"] + "/done"),
            trace_observable("done", node),
            ops.share(),
        ).subscribe(s["done"])
        reset_disp.add(d)

        d = msgs.pipe(
            filter_dict_on_key(s["name"]),
            ops.filter(lambda msg: msg is not None),
            ops.map(s["converter"].convert),
            ops.share(),
        ).subscribe(s["msg"])
        reset_disp.add(d)

    ###########################################################################
    # Reset ###################################################################
    ###########################################################################
    R = Subject()
    reset = dict(name="reset", address=ns + "/reset", msg_type=UInt64, msg=R)

    # Prepare node reset topic
    node_reset = dict(
        name=node.ns_name,
        address=node.ns_name + "/end_reset",
        msg_type=Bool,
        msg=Subject(),
    )

    # Reset pipeline
    d = R.pipe(ops.map(lambda x: Bool(data=True))).subscribe(node_reset["msg"], scheduler=tp_scheduler)
    reset_disp.add(d)

    ###########################################################################
    # Register ################################################################
    ###########################################################################
    REG_OBJECT = Subject()  # ---> Not a node output, but used in node.register_object() to kickstart register pipeline.
    register_object = dict(
        name="register_object",
        address=ns + "/register_object",
        msg=Subject(),
        msg_type=String,
    )
    REG_NODE = Subject()  # ---> Not a node output, but used in node.register_node() to kickstart register pipeline.
    register_node = dict(
        name="register_node",
        address=ns + "/register_node",
        msg=Subject(),
        msg_type=String,
    )

    # Register pipeline
    d = REG_OBJECT.subscribe(register_object["msg"], scheduler=tp_scheduler)
    reset_disp.add(d)
    d = REG_NODE.subscribe(register_node["msg"], scheduler=tp_scheduler)
    reset_disp.add(d)

    ###########################################################################
    # End reset ###############################################################
    ###########################################################################
    tick = dict(name="tick", address=ns + "/bridge/outputs/tick", msg=Subject(), msg_type=UInt64)
    end_reset = dict(name="end_reset", address=ns + "/end_reset", msg=Subject(), msg_type=UInt64)
    d = end_reset["msg"].pipe(spy("RESET END", node, log_level=DEBUG)).subscribe(tick["msg"])
    reset_disp.add(d)

    # Create node inputs & outputs
    node_inputs = [reset, end_reset, end_register]
    node_outputs = [register_object, register_node, start_reset, tick, node_reset, real_reset]
    outputs = []

    # Create return objects
    env_subjects = dict(register_object=REG_OBJECT, register_node=REG_NODE, start_reset=SR)
    rx_objects = dict(
        node_inputs=node_inputs,
        node_outputs=node_outputs,
        outputs=outputs,
        state_outputs=state_outputs + tuple(done_outputs),
        disposable=reset_disp,
    )
    return rx_objects, env_subjects
